from functools import partial
import networkx as nx
import scipy
import scipy.special
from scipy.special import comb
import scipy.sparse
import autograd.numpy as np
import msprime
from .compute_sfs import expected_total_branch_len
from .data.compressed_counts import _CompressedHashedCounts, _CompressedList
from .data.snps import SnpAlleleCounts
from .util import memoize_instance
from .math_functions import (
    binom_coeffs, roll_axes, hypergeom_quasi_inverse,
    par_einsum, convolve_sum_axes)

import pysam
import os
import itertools

import logging
logger = logging.getLogger(__name__)


class differentiable_method(object):
    """
    a descriptor for cacheing all the differentiable objects in the demography
    this is used to reorganize some of the computations during automatic differentiation,
    which can be very resource intensive

    based on memoize_instance in util.py, which is itself based on http://code.activestate.com/recipes/577452-a-memoize-decorator-for-instance-methods/
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self.func
        return partial(self, obj)

    def __call__(self, *args, **kw):
        obj = args[0]
        cache = obj._diff_cache

        key = (self.func, args[1:], frozenset(list(kw.items())))
        try:
            res = cache[key]
        except KeyError:
            res = cache[key] = self.func(*args, **kw)
        return res

class Demography(object):
    """
    The demographic history relating a sample of individuals.
    """
    def __init__(self, G, cache=None):
        """
        For internal use only.
        Use make_demography() to create a Demography.
        """
        self._G = G
        self._event_tree = _build_event_tree(self._G)

        if cache is not None:
            self._diff_cache = cache
        else:
            self._diff_cache = {}

    def _get_differentiable_part(self):
        # used with self._get_graph_structure() and autograd.checkpoint
        # returns a dict of the memoized values so we can
        # compute their derivatives easily
        expected_total_branch_len(self)
        return self._diff_cache

    def _get_graph_structure(self):
        # returns just the graph structure, i.e. the "non-differentiable" part of the Demography
        # use this with _get_differentiable_part()
        # to re-organize certain computations during automatic differentiation
        ret = nx.DiGraph()
        ret.add_nodes_from(self._G.nodes(data=False))
        ret.add_edges_from(self._G.edges(data=False))

        for v, d in self._G.nodes(data=True):
            if 'lineages' in d:
                ret.nodes[v]['lineages'] = d['lineages']

        ret.graph['events_as_edges'] = tuple(self._G.graph['events_as_edges'])
        ret.graph['sampled_pops'] = self.sampled_pops

        return ret

    #def copy(self, sampled_n=None):
    #    """
    #    Notes
    #    -----
    #    Note that momi.expected_sfs, momi.composite_log_likelihood require
    #    Demography.sampled_n == ConfigList.sampled_n.
    #    If this is not the case, you can use copy() to create a copy with the correct
    #    sampled_n.
    #    """
    #    if sampled_n is None:
    #        sampled_n = self.sampled_n
    #    return _make_multipop_moran(self.events, self.sampled_pops, sampled_n, self.sampled_t, self.default_N)

    #@property
    #def events(self):
    #    """
    #    The list of events (tuples) making up the demographic history
    #    """
    #    return self._G.graph['event_cmds']

    @property
    def sampled_pops(self):
        """
        The list of population labels
        """
        return self._G.graph['sampled_pops']

    @property
    def sampled_n(self):
        """
        The list of number of samples per population
        """
        return np.array(tuple(self._G.nodes[(l, 0)]['lineages'] for l in self.sampled_pops), dtype=int)

    @memoize_instance
    def _n_at_node(self, node):
        return sum(self._G.nodes[(pop, idx)]['lineages']
                   for pop, idx in nx.dfs_preorder_nodes(self._G, node)
                   if idx == 0)

    @property
    def _root(self):
        ret, = self._parent_pops(self._event_root)
        return ret

    @property
    def _event_root(self):
        return self._event_tree.root

    def _event_type(self, event):
        if len(event) == 1:
            return 'leaf'
        elif len(event) == 3:
            return 'pulse'
        elif len(self._event_tree[event]) == 2:
            return 'merge_clusters'
        else:
            return 'merge_subpops'

    def _sub_pops(self, event):
        '''
        The group of subpopulations corresponding to this event in the junction tree.
        '''
        return self._event_tree.nodes[event]['subpops']

    def _parent_pops(self, event):
        '''The populations arising due to this event, backwards in time.'''
        return self._event_tree.nodes[event]['parent_pops']

    def _child_pops(self, event):
        '''
        Returns dict of
        {child_pop : child_event},
        which gives populations arising from this event forward in time,
        and the corresponding child events in the junction tree.
        '''
        return self._event_tree.nodes[event]['child_pops']

    def _pulse_nodes(self, event):
        parent_pops = self._parent_pops(event)
        child_pops_events = self._child_pops(event)
        assert len(child_pops_events) == 2
        child_pops, child_events = list(zip(*list(child_pops_events.items())))

        child_in = dict(self._G.in_degree(child_pops))
        recipient, = [k for k, v in list(child_in.items()) if v == 2]
        non_recipient, = [k for k, v in list(child_in.items()) if v == 1]

        parent_out = dict(self._G.out_degree(parent_pops))
        donor, = [k for k, v in list(parent_out.items()) if v == 2]
        non_donor, = [k for k, v in list(parent_out.items()) if v == 1]

        return recipient, non_recipient, donor, non_donor

    """
    ALL differentiable methods used by compute_sfs
    should be decorated by @differentiable_method !!!

    This is so that we can extract the memoized values with
    _get_differentiable_part(), which can be used in conjunction with
    autograd.checkpoint() to reduce memory usage by computing
    in batches of SNPs
    """

    @property
    @differentiable_method
    def sampled_t(self):
        """
        An array of times at which each population was sampled
        """
        return np.array(tuple(self._G.nodes[(l, 0)]['sizes'][0]['t'] for l in self.sampled_pops))

    @property
    @differentiable_method
    def default_N(self):
        """
        The scaled size N of all populations, unless changed by -en or -eg
        """
        return self._G.graph['default_N']

    @differentiable_method
    def _truncated_sfs(self, node):
        return self._G.nodes[node]['model'].sfs(self._n_at_node(node))

    @differentiable_method
    def _scaled_time(self, node):
        return self._G.nodes[node]['model'].scaled_time

    def _pulse_prob(self, event):
        return self._pulse_prob_helper(event), self._pulse_prob_idxs(event)

    def _pulse_prob_idxs(self, event):
        recipient, non_recipient, donor, non_donor = self._pulse_nodes(event)
        admixture_idxs = self._admixture_prob_idxs(recipient)
        return admixture_idxs + [non_recipient]

    @differentiable_method
    def _pulse_prob_helper(self, event):
        # returns 4-tensor
        # running time is O(n^5), because of pseudo-inverse
        # if pulse from ghost population, only costs O(n^4)
        recipient, non_recipient, donor, non_donor = self._pulse_nodes(event)

        admixture_prob, admixture_idxs = self._admixture_prob(recipient)

        pulse_idxs = admixture_idxs + [non_recipient]
        assert pulse_idxs == self._pulse_prob_idxs(event)

        pulse_prob = par_einsum(admixture_prob, admixture_idxs,
                                binom_coeffs(self._n_at_node(non_recipient)), [
                                    non_recipient],
                                pulse_idxs)
        pulse_prob = par_einsum(pulse_prob, pulse_idxs,
                                binom_coeffs(self._n_at_node(recipient)), [
                                    donor],
                                pulse_idxs)
        pulse_prob = roll_axes(pulse_prob, pulse_idxs, non_recipient, donor)

        donor_idx = pulse_idxs.index(donor)
        pulse_prob = par_einsum(pulse_prob, pulse_idxs,
                                1.0 /
                                binom_coeffs(pulse_prob.shape[
                                             donor_idx] - 1), [donor],
                                pulse_idxs)

        # reduce the number of lineages in donor to only the number necessary
        N, n = pulse_prob.shape[donor_idx] - 1, self._n_at_node(donor)
        assert N >= n
        if N > n:
            assert -1 not in pulse_idxs
            tmp_idxs = [-1 if x == donor else x for x in pulse_idxs]
            pulse_prob = par_einsum(pulse_prob, tmp_idxs,
                                    hypergeom_quasi_inverse(N, n),
                                    [-1, donor], pulse_idxs)
        assert pulse_prob.shape[donor_idx] == n + 1

        return pulse_prob

    def _admixture_prob(self, admixture_node):
        return self._admixture_prob_helper(admixture_node), self._admixture_prob_idxs(admixture_node)

    def _admixture_prob_idxs(self, admixture_node):
        edge1, edge2 = sorted(self._G.in_edges(
            [admixture_node], data=True), key=lambda x: str(x[:2]))
        parent1, parent2 = [e[0] for e in (edge1, edge2)]
        return [admixture_node, parent1, parent2]

    @differentiable_method
    def _admixture_prob_helper(self, admixture_node):
        '''
        Array with dim [n_admixture_node+1, n_parent1_node+1, n_parent2_node+1],
        giving probability of derived counts in child, given derived counts in parents
        '''
        n_node = self._n_at_node(admixture_node)

        # admixture node must have two parents
        edge1, edge2 = sorted(self._G.in_edges(
            [admixture_node], data=True), key=lambda x: str(x[:2]))
        parent1, parent2 = [e[0] for e in (edge1, edge2)]
        prob1, prob2 = [e[2]['prob'] for e in (edge1, edge2)]
        assert prob1 + prob2 == 1.0

        #n_from_1 = np.arange(n_node + 1)
        #n_from_2 = n_node - n_from_1
        #binom_coeffs = (prob1**n_from_1) * (prob2**n_from_2) * \
        #    scipy.special.comb(n_node, n_from_1)
        #ret = par_einsum(_der_in_admixture_node(n_node), list(range(4)),
        #                 binom_coeffs, [0],
        #                 [1, 2, 3])
        ret = np.transpose(admixture_operator(n_node, prob1))
        assert ret.shape == tuple([n_node + 1] * 3)

        assert [admixture_node, parent1,
                parent2] == self._admixture_prob_idxs(admixture_node)
        return ret

    def simulate_data(self, length, num_replicates=1, **kwargs):
        treeseq = self.simulate_trees(length=length, num_replicates=num_replicates,
                                      **kwargs)
        try:
            treeseq.variants
        except:
            pass
        else:
            treeseq = [treeseq]

        mat = np.zeros((len(self.sampled_n), sum(self.sampled_n)), dtype=int)
        j = 0
        for i, n in enumerate(self.sampled_n):
            for _ in range(n):
                mat[i, j] = 1
                j += 1
        mat = scipy.sparse.csr_matrix(mat)

        def get_config(genos):
            derived_counts = mat.dot(genos)
            return np.array([
                self.sampled_n - derived_counts,
                derived_counts
            ]).T

        #chrom = []
        chrom = _CompressedList()
        pos = []
        compressed_counts = _CompressedHashedCounts(len(self.sampled_pops))

        for c, locus in enumerate(treeseq):
            for v in locus.variants():
                compressed_counts.append(get_config(v.genotypes))
                chrom.append(c)
                pos.append(v.position)

        return SnpAlleleCounts(
            chrom, pos, compressed_counts.compressed_allele_counts(),
            self.sampled_pops, use_folded_sfs=False,
            non_ascertained_pops=[], length=length*num_replicates,
            n_read_snps=len(compressed_counts), n_excluded_snps=0)

    def simulate_vcf(self, out_prefix, mutation_rate,
                     recombination_rate, length,
                     chrom_name=1, ploidy=1, random_seed=None,
                     force=False, print_aa=True):
        out_prefix = os.path.expanduser(out_prefix)
        vcf_name = out_prefix + ".vcf"
        bed_name = out_prefix + ".bed"
        for fname in (vcf_name, bed_name):
            if not force and os.path.isfile(fname):
                raise FileExistsError(
                    "{} exists and force=False".format(fname))

        if np.any(self.sampled_n % ploidy != 0):
            raise ValueError("Sampled alleles per population must be"
                             " integer multiple of ploidy")

        with open(bed_name, "w") as bed_f:
            print(chrom_name, 0, length, sep="\t", file=bed_f)

        with open(vcf_name, "w") as vcf_f:
            treeseq = self.simulate_trees(
                mutation_rate=mutation_rate,
                recombination_rate=recombination_rate,
                length=length, num_replicates=1,
                random_seed=random_seed)

            print("##fileformat=VCFv4.2", file=vcf_f)
            print('##source="VCF simulated by momi2 using'
                  ' msprime backend"', file=vcf_f)
            print("##contig=<ID={chrom_name},length={length}>".format(
                chrom_name=chrom_name, length=length), file=vcf_f)
            print('##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">',
                  file=vcf_f)
            print('##INFO=<ID=AA,Number=1,Type=String,Description="Ancestral Allele">',
                  file=vcf_f)

            n_samples = int(np.sum(self.sampled_n) / ploidy)
            fields = ["#CHROM", "POS", "ID", "REF", "ALT", "QUAL",
                      "FILTER", "INFO", "FORMAT"]
            for pop, n in zip(self.sampled_pops, self.sampled_n):
                for i in range(int(n / ploidy)):
                    fields.append("{}_{}".format(pop, i))
            print(*fields, sep="\t", file=vcf_f)

            loc = next(treeseq)
            if print_aa:
                info_str = "AA=A"
            else:
                info_str = "."

            for v in loc.variants():
                gt = np.reshape(v.genotypes, (n_samples, ploidy))
                print(chrom_name, int(np.floor(v.position)),
                      ".", "A", "T", ".", ".", info_str, "GT",
                      *["|".join(map(str, sample)) for sample in gt],
                      sep="\t", file=vcf_f)

        pysam.tabix_index(vcf_name, preset="vcf", force=force)

    def simulate_trees(self, **kwargs):
        sampled_t = self.sampled_t
        if sampled_t is None:
            sampled_t = 0.0
        sampled_t = np.array(sampled_t) * np.ones(len(self.sampled_pops))

        pops = {p: i for i, p in enumerate(self.sampled_pops)}

        demographic_events = []
        for e in self._G.graph["events"]:
            e = e.get_msprime_event(self._G.graph["params"], pops)
            if e is not None:
                demographic_events.append(e)

        return msprime.simulate(
            population_configurations=[
                msprime.PopulationConfiguration()
                for _ in range(len(pops))],
            Ne=self.default_N / 4,
            demographic_events=demographic_events,
            samples=[
                msprime.Sample(population=pops[p], time=t)
                for p, t, n in zip(
                        self.sampled_pops, self.sampled_t,
                        self.sampled_n)
                for _ in range(n)],
            **kwargs)


def rescale_events(events, factor):
    rescaled_events = []
    for event in events:
        if event[0] == '-ej':
            flag, t, i, j = event
            event = (flag, t * factor, i, j)
        elif event[0] == '-en':
            flag, t, i, N = event
            event = (flag, t * factor, i, N * factor)
        elif event[0] == '-eg':
            flag, t, i, alpha = event
            event = (flag, t * factor, i, alpha / (1.0 * factor))
        elif event[0] == '-ep':
            flag, t, i, j, p = event
            event = (flag, t * factor, i, j, p)
        else:
            assert False
        rescaled_events += [event]
    return rescaled_events


def admixture_operator(n_node, p):
    # axis0=n_from_parent, axis1=der_from_parent, axis2=der_in_parent
    der_in_parent = np.tile(np.arange(n_node + 1), (n_node + 1, n_node + 1, 1))
    n_from_parent = np.transpose(der_in_parent, [2, 0, 1])
    der_from_parent = np.transpose(der_in_parent, [0, 2, 1])

    anc_in_parent = n_node - der_in_parent
    anc_from_parent = n_from_parent - der_from_parent

    x = comb(der_in_parent, der_from_parent) * comb(
        anc_in_parent, anc_from_parent) / comb(n_node, n_from_parent)
    # rearrange so axis0=1, axis1=der_in_parent, axis2=der_from_parent, axis3=n_from_parent
    x = np.transpose(x)
    x = np.reshape(x, [1] + list(x.shape))

    n = np.arange(n_node+1)
    B = comb(n_node, n)

    # the two arrays to convolve_sum_axes
    x1 = (x * B * ((1-p)**n) * (p**(n[::-1])))
    x2 = x[:, :, :, ::-1]

    # reduce array size; approximate low probability events with 0
    mu = n_node * (1-p)
    sigma = np.sqrt(n_node * p * (1-p))
    n_sd = 4
    lower = np.max((0, np.floor(mu - n_sd*sigma)))
    upper = np.min((n_node, np.ceil(mu + n_sd*sigma))) + 1
    lower, upper = int(lower), int(upper)

    ##x1 = x1[:,:,:upper,lower:upper]
    ##x2 = x2[:,:,:(n_node-lower+1),lower:upper]

    ret = convolve_sum_axes(x1, x2)
    # axis0=der_in_parent1, axis1=der_in_parent2, axis2=der_in_child
    ret = np.reshape(ret, ret.shape[1:])
    if ret.shape[2] < (n_node+1):
        ret = np.pad(ret, [(0,0),(0,0),(0,n_node+1-ret.shape[2])], "constant")
    return ret[:, :, :(n_node+1)]


#@memoize
#def _der_in_admixture_node(n_node):
#    '''
#    returns 4d-array, [n_from_parent1, der_in_child, der_in_parent1, der_in_parent2].
#    Used by Demography._admixture_prob_helper
#    '''
#    # axis0=n_from_parent, axis1=der_from_parent, axis2=der_in_parent
#    der_in_parent = np.tile(np.arange(n_node + 1), (n_node + 1, n_node + 1, 1))
#    n_from_parent = np.transpose(der_in_parent, [2, 0, 1])
#    der_from_parent = np.transpose(der_in_parent, [0, 2, 1])
#
#    anc_in_parent = n_node - der_in_parent
#    anc_from_parent = n_from_parent - der_from_parent
#
#    x = scipy.special.comb(der_in_parent, der_from_parent) * scipy.special.comb(
#        anc_in_parent, anc_from_parent) / scipy.special.comb(n_node, n_from_parent)
#
#    ret, labels = convolve_axes(
#        x, x[::-1, ...], [[c for c in 'ijk'], [c for c in 'ilm']], ['j', 'l'], 'n')
#    return np.einsum('%s->inkm' % ''.join(labels), ret[..., :(n_node + 1)])


def _build_event_tree(G):
    # def node_time(v):
    #     return G.nodes[v]['sizes'][0]['t']

    eventEdgeList = []
    currEvents = {k: (k,) for k, v in list(dict(G.out_degree()).items()) if v == 0}
    eventDict = {e: {'subpops': (v,), 'parent_pops': (
        v,), 'child_pops': {}} for v, e in list(currEvents.items())}
    for e in G.graph['events_as_edges']:
        # get the population edges forming the event
        parent_pops, child_pops = list(map(set, list(zip(*e))))
        child_events = set([currEvents[c] for c in child_pops])

        sub_pops = set(itertools.chain(
            *[eventDict[c]['subpops'] for c in child_events]))
        sub_pops.difference_update(child_pops)
        sub_pops.update(parent_pops)

        # try:
        #     times = [t for t in map(node_time, parent_pops)]
        #     assert np.allclose(times, times[0])
        # except TypeError:
        #     ## autograd sometimes raise TypeError for this assertion
        #     pass

        eventDict[e] = {'parent_pops': tuple(parent_pops), 'subpops': tuple(
            sub_pops), 'child_pops': {c: currEvents[c] for c in child_pops}}
        currEvents.update({p: e for p in sub_pops})
        for p in child_pops:
            del currEvents[p]
        eventEdgeList += [(e, c) for c in child_events]
    ret = nx.DiGraph(eventEdgeList)
    for e in eventDict:
        ret.add_node(e, **(eventDict[e]))

    assert len(currEvents) == 1
    root, = [v for k, v in list(currEvents.items())]
    ret.root = root

    return ret

# methods for constructing demography from string



def get_treeseq_configs(treeseq, sampled_n):
    mat = np.zeros((len(sampled_n), sum(sampled_n)), dtype=int)
    j = 0
    for i, n in enumerate(sampled_n):
        for _ in range(n):
            mat[i, j] = 1
            j += 1
    mat = scipy.sparse.csr_matrix(mat)

    def get_config(genos):
        derived_counts = mat.dot(genos)
        return np.array([
            sampled_n - derived_counts,
            derived_counts
        ]).T

    for v in treeseq.variants():
        yield get_config(v.genotypes)




import autograd.numpy as np
from functools import partial, wraps
#from autograd.core import primitive, Node
from autograd.extend import primitive, defvjp


def count_calls(fun):
    call_counter = [0]

    @wraps(fun)
    def new_fun(*args, **kwargs):
        call_counter[0] += 1
        return fun(*args, **kwargs)

    new_fun.num_calls = lambda: call_counter[0]

    def reset_count(): call_counter[0] = 0
    new_fun.reset_count = reset_count

    return new_fun


def check_symmetric(X):
    Xt = np.transpose(X)
    assert np.allclose(X, Xt)
    return 0.5 * (X + Xt)


def check_psd(X, **tol_kwargs):
    X = check_symmetric(X)
    d, U = np.linalg.eigh(X)
    d = truncate0(d, **tol_kwargs)
    ret = np.dot(U, np.dot(np.diag(d), np.transpose(U)))
    #assert np.allclose(ret, X)
    return np.array(ret, ndmin=2)


def truncate0(x, axis=None, strict=False, tol=1e-13):
    '''make sure everything in x is non-negative'''
    # the maximum along axis
    maxes = np.maximum(np.amax(x, axis=axis), 1e-300)
    # the negative part of minimum along axis
    mins = np.maximum(-np.amin(x, axis=axis), 0.0)

    # assert the negative numbers are small (relative to maxes)
    assert np.all(mins <= tol * maxes)

    if axis is not None:
        idx = [slice(None)] * x.ndim
        idx[axis] = np.newaxis
        mins = mins[idx]
        maxes = maxes[idx]

    if strict:
        # set everything below the tolerance to 0
        return set0(x, x < tol * maxes)
    else:
        # set everything of same magnitude as most negative number, to 0
        return set0(x, x < 2 * mins)


def check_probs_matrix(x):
    x = truncate0(x)
    rowsums = np.sum(x, axis=1)
    assert np.allclose(rowsums, 1.0)
    return np.einsum('ij,i->ij', x, 1.0 / rowsums)


@primitive
def set0(x, indices):
    y = np.array(x)
    y[indices] = 0
    return y
defvjp(set0, lambda ans, x, indices: lambda g: set0(g, indices))
#set0.defgrad(lambda ans, x, indices: lambda g: set0(g, indices))
#set0.defvjp(lambda g, ans, vs, gvs, x, indices: set0(g, indices))


def closeleq(x, y):
    return np.logical_or(np.isclose(x, y), x <= y)


def closegeq(x, y):
    return np.logical_or(np.isclose(x, y), x >= y)


@primitive
def make_constant(x):
    return x
#make_constant.defgrad(lambda ans, x: lambda g: np.zeros(x.shape, dtype=x.dtype))
#make_constant.defvjp(lambda g, ans, vs, gvs, x: np.zeros(
#    x.shape, dtype=x.dtype))
defvjp(make_constant, lambda ans, x: lambda g: np.zeros(x.shape, dtype=x.dtype))


def memoize(obj):
    cache = obj.cache = {}

    @wraps(obj)
    def memoizer(*args, **kwargs):
        if args not in cache:
            cache[args] = obj(*args, **kwargs)
        return cache[args]
    return memoizer


class memoize_instance(object):
    """cache the return value of a method

    This class is meant to be used as a decorator of methods. The return value
    from a given method invocation will be cached on the instance whose method
    was invoked. All arguments passed to a method decorated with memoize must
    be hashable.

    If a memoized method is invoked directly on its class the result will not
    be cached. Instead the method will be invoked like a static method:
    class Obj(object):
        @memoize
        def add_to(self, arg):
            return self + arg
    Obj.add_to(1) # not enough arguments
    Obj.add_to(1, 2) # returns 3, result is not cached

    recipe from http://code.activestate.com/recipes/577452-a-memoize-decorator-for-instance-methods/
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self.func
        return partial(self, obj)

    def __call__(self, *args, **kw):
        obj = args[0]
        try:
            cache = obj.__cache
        except AttributeError:
            cache = obj.__cache = {}
        key = (self.func, args[1:], frozenset(list(kw.items())))
        try:
            res = cache[key]
        except KeyError:
            res = cache[key] = self.func(*args, **kw)
        return res




import momi
from momi import expected_sfs, expected_sfs_tensor_prod, sfs_tensor_prod
import pytest
import random
import autograd.numpy as np
import scipy
import scipy.stats
import itertools
import sys

from demo_utils import simple_admixture_demo, random_tree_demo


def check_random_tensor(demo, *args, **kwargs):
    leaves = demo.sampled_pops
    ranges = [list(range(n + 1)) for n in demo.sampled_n]

    config_list = momi.data.configurations.build_full_config_list(demo.sampled_pops, demo.sampled_n)

    esfs = expected_sfs(demo, config_list, *args, **kwargs)

    tensor_components = [np.random.normal(size=(1, n + 1)) for n in demo.sampled_n]
    #tensor_components_list = tuple(v[0,:] for _,v in sorted(tensor_components.iteritems()))

    #prod1 = sfs_tensor_prod(dict(list(zip(config_list,esfs))), tensor_components)
    # sfs = momi.site_freq_spectrum(demo.sampled_pops, [dict(zip((tuple(map(tuple,c)) for c in config_list),
    #                                                           esfs))])
    sfs = momi.site_freq_spectrum(demo.sampled_pops,
                                  [{tuple(map(tuple, c)): s for c, s in zip(config_list, esfs)}])
    #assert sfs.get_dict() == {tuple(map(tuple,c)): s for c,s in zip(config_list, esfs)}
    prod1 = sfs_tensor_prod(sfs, tensor_components)
    # prod1 = sfs_tensor_prod({tuple(map(tuple,c)): s for c,s in zip(config_list, esfs)},
    #                        tensor_components)
    prod2 = expected_sfs_tensor_prod(
        tensor_components, demo, sampled_pops=demo.sampled_pops)

    assert np.allclose(prod1, prod2)


def test_tree_demo_rank1tensor():
    lins_per_pop = 2
    num_leaf_pops = 3

    demo = random_tree_demo(num_leaf_pops, lins_per_pop)
    check_random_tensor(demo._get_demo(dict(zip(demo.leafs, [lins_per_pop]*num_leaf_pops))))


def test_admixture_demo_rank1tensor():
    demo = simple_admixture_demo()
    check_random_tensor(demo._get_demo({"a":4,"b":5}))



import heapq as hq
import collections as co
import autograd.numpy as np
import matplotlib as mpl
from matplotlib import pyplot as plt
from matplotlib.ticker import LogFormatterSciNotation
import logging


class DemographyPlot(object):
    """Object for plotting a demography.

    After creating this object, call :meth:`DemographyPlot.draw`
    to draw the demography.

    :param DemographicModel model: \
    model to plot

    :param list,dict pop_x_positions: \
    list ordering populations along the x-axis, or dict mapping \
    population names to x-axis positions

    :param matplotlib.axes.Axes ax: \
    Canvas to draw the figure on. Defaults to \
    :func:`matplotlib.pyplot.gca`

    :param tuple figsize: \
    If non-None, calls :func:`matplotlib.pyplot.figure` to \
    create a new figure with this width and height. \
    Ignored if ``ax`` is non-None.

    :param float linthreshy: \
    Scale y-axis linearly below this value \
    and logarithmically above it. (Default: use linear scaling only)

    :param list minor_yticks:  Minor ticks on y axis
    :param list major_yticks: Major ticks on y axis
    :param str,matplotlib.colors.Colormap color_map: \
    Colormap mapping pulse probabilities to colors

    :param tuple pulse_color_bounds: pair of ``(lower, upper)`` bounds \
    for mapping pulse probabilities to colors
    """
    def __init__(self, model, pop_x_positions,
                 ax=None, figsize=None,
                 linthreshy=None, minor_yticks=None,
                 major_yticks=None,
                 color_map="cool",
                 pulse_color_bounds=(0, 1),
                 draw=True):
        self.leafs = model.leafs
        self.model = model.copy()
        try:
            pop_x_positions.items()
        except AttributeError:
            pop_x_positions = {p: i for i, p in enumerate(
                pop_x_positions)}
        self.x_pos = dict(pop_x_positions)
        if ax:
            self.ax = ax
        else:
            if figsize:
                plt.figure(figsize=figsize)
            fig = plt.gcf()
            fig.clf()
            self.ax = fig.gca()

        if linthreshy:
            self.ax.set_yscale('symlog', linthresh=linthreshy)
            self.ax.get_yaxis().set_major_formatter(
                LogFormatterSciNotation(labelOnlyBase=False,
                                        minor_thresholds=(100, 100),
                                        linthresh=5e4))
            self.ax.axhline(y=linthreshy, linestyle=":",
                            color="black", zorder=1)

        self.minor_yticks = minor_yticks
        self.major_yticks = major_yticks
        self.additional_times = []
        if minor_yticks:
            self.additional_times.extend(minor_yticks)
        if major_yticks:
            self.additional_times.extend(major_yticks)
        self.additional_times = sorted(self.additional_times)

        self._init_plot(model)

        self.all_N = [p.N for popline in self._plot.pop_lines.values()
                      for p in popline.points]
        self.base_N = min(self.all_N)
        self.pmin, self.pmax = pulse_color_bounds
        self.cm_scalar_mappable = plt.cm.ScalarMappable(
            norm=mpl.colors.Normalize(vmin=self.pmin, vmax=self.pmax),
            cmap=color_map)

        self.cbar = None
        if draw:
            self.draw()

    def _init_plot(self, model):
        self._plot = _DemographyPlot(
            self.x_pos, model.N_e, self.additional_times)

        params_dict = model.get_params()
        event_list = sorted(list(model.leaf_events) +
                            list(model.size_events) +
                            list(model.topology_events),
                            key=lambda e: e.t(params_dict))

        for e in event_list:
            e.add_to_plot(params_dict, self._plot)

        for pop in self._plot.pop_lines.values():
            pop.goto_time(float('inf'))

    def draw(self, alpha=1.0, tree_color="C0", draw_frame=True,
             rad=0, pulse_label=True):
        """Draw the demography.

        This is method draws the demography by calling
        :meth:`DemographyPlot.draw_tree`,
        :meth:`DemographyPlot.draw_leafs`,
        :meth:`DemographyPlot.draw_pulse`, and
        :meth:`DemographyPlot.draw_frame`.
        Call those methods directly for finer control.

        :param float alpha: Level of transparency (0=transparent, 1=opaque)
        :param str tree_color: Color of tree
        :param bool draw_frame: If True, call \
        :meth:`DemographyPlot.draw_frame` to draw tickmarks and legends

        :param float rad: Arc of pulse arrows in radians.
        :param bool pulse_label: If True, label each pulse with its strength
        """
        self.draw_tree(alpha=alpha, tree_color=tree_color)
        self.draw_leafs(alpha=alpha)
        for pop1, pop2, t, p in self.iter_pulses():
            self.draw_pulse(pop1, pop2, t, p, alpha=alpha,
                            rad=rad, pulse_label=pulse_label)
        if draw_frame:
            self.draw_frame()

    def draw_frame(self, pops=None, rename_pops=None, rotation=-30):
        """Draw tickmarks, legend, and colorbar.

        :param list pops: Populations to label on x-axis. \
        If None, add all populations

        :param dict rename_pops: Dict to give new names to \
        populations on x-axis.

        :param float rotation: Degrees to rotate x-axis population labels by.
        """
        self.draw_xticks(pops=pops, rename_pops=rename_pops,
                         rotation=rotation)
        self.draw_N_legend()
        self.draw_pulse_colorbar()
        if self.minor_yticks:
            self.ax.set_yticks(self.minor_yticks, minor=True)
        if self.major_yticks:
            self.ax.set_yticks(self.major_yticks)
        self.ax.set_ylabel("Time")

    def draw_pulse_colorbar(self):
        self.cm_scalar_mappable.set_array([])
        if not self.cbar:
            self.cbar = self.ax.get_figure().colorbar(
                self.cm_scalar_mappable, fraction=0.046, pad=0.04,
                ax=self.ax)
            self.cbar.set_label("Pulse Probability")

    def add_bootstrap(self, params, alpha,
                      rad=-.1, rand_rad=True):
        """Add an inferred bootstrap demography to the plot.

        :param dict params: Inferred bootstrap parameters
        :param float alpha: Transparency
        :param float rad: Arc of pulse arrows in radians
        :param bool rand_rad: Add random jitter to the arc of the pulse arrows
        """
        model = self.model.copy()
        model.set_params(params)
        additional_plot = AdditionalDemographyPlot(self, model)
        if rand_rad:
            rad *= 2 * np.random.uniform()
        additional_plot.draw(alpha=alpha, tree_color="gray",
                             pulse_label=False, rad=rad)

    def draw_xticks(self, pops=None, rename_pops=None, rotation=-30):
        if pops is None:
            pops = list(self._plot.pop_lines.keys())
        if rename_pops is None:
            rename_pops = {}
        x_pos = {rename_pops.get(p, p): self.x_pos[p] for p in pops}
        xtick_labs, xticks = zip(*sorted(
            x_pos.items(), key=lambda itm: itm[1]))

        self.ax.set_xticks(xticks)
        self.ax.set_xticklabels(xtick_labs,
                                rotation=rotation)

    def draw_leafs(self, leafs=None,
                   facecolors="none", edgecolors="black",
                   s=100, zorder=2, **kwargs):
        """Draw symbols for the leaf populations.

        :param list leafs: Leaf populations to draw symbols for. \
        If None, add all leafs.

        Extra parameters ``facecolors, edgecolors, s, zorder, **kwargs``
        are passed to :meth:`matplotlib.axes.Axes.scatter`.
        """
        if leafs is None:
            leafs = self.leafs

        x = []
        t = []
        for l in leafs:
            popline = self._plot.pop_lines[l]
            for p in popline.points:
                if p.is_leaf:
                    x.append(self.x_pos[l])
                    t.append(p.t)
                    break

        self.ax.scatter(x, t, facecolors=facecolors,
                        edgecolors=edgecolors,
                        s=s, zorder=zorder, **kwargs)

    def N_to_linewidth(self, N):
        return np.log(N/self.base_N) + 2

    def draw_tree(self, tree_color="C0", alpha=1.0):
        """Draw the demographic tree (without pulse migrations)

        :param str tree_color: Color of the tree
        :param float alpha: Transparency

        """
        # draw vertical lines
        for popname, popline in self._plot.pop_lines.items():
            x = [self.x_pos[popname]]*len(popline.points)
            t = [p.t for p in popline.points]
            xt = list(zip(x, t))

            is_ghost = [False for _ in popline.points]
            for i, p in enumerate(popline.points):
                if p.is_leaf:
                    for j in range(i):
                        is_ghost[j] = True

            for bottom, top, N, ghost in zip(
                    xt[:-1], xt[1:],
                    [p.N for p in popline.points[:-1]],
                    is_ghost):
                curr_x, curr_t = zip(bottom, top)

                if ghost:
                    linestyle = ":"
                else:
                    linestyle = "-"

                self.ax.plot(
                    curr_x, curr_t, color=tree_color,
                    linewidth=self.N_to_linewidth(N),
                    linestyle=linestyle,
                    alpha=alpha, zorder=1)

        # draw horizontal lines
        for arrow in self._plot.pop_arrows:
            if arrow.p != 1:
                continue
            self.ax.plot(
                (arrow.from_pop.x, arrow.to_pop.x),
                (arrow.t, arrow.t), color=tree_color,
                linewidth=self.N_to_linewidth(arrow.from_N),
                alpha=alpha)

    def iter_pulses(self):
        """Iterate over the pulse arrows, to pass to :meth:`DemographyPlot.draw_pulse`

        :returns: iterator over tuples ``(pop1, pop2, t, p)``
        """
        for arrow in self._plot.pop_arrows:
            if arrow.p != 1:
                yield (arrow.to_pop.name, arrow.from_pop.name,
                       arrow.t, arrow.p)

    def draw_pulse(self, pop1, pop2, t, p,
                   rad=-.1, alpha=1.0,
                   pulse_label=True, adj_label_x=0, adj_label_y=0):
        """Draw a pulse.

        Use :meth:`DemographyPlot.iter_pulses` to iterate over \
        the pulses, which can then be plotted with this method.

        :param str pop1: Population the arrow is pointing into
        :param str pop2: Population the arrow is pointing away from
        :param float t: Time of the pulse
        :param float p: Strength of the pulse
        :param float rad: Arc of the pulse arrow in radians.
        :param float alpha: Transparency
        :param bool pulse_label: Add a label for pulse strength?
        :param float adj_label_x: Adjust pulse label x position
        :param float adj_label_y: Adjust pulse label y position
        """
        if p < self.pmin or p > self.pmax:
            logging.getLogger(__name__).warn("Trying to plot pulse p {} not in {}".format(p, (self.pmin, self.pmax)))

        frm = self.x_pos[pop2]
        to = self.x_pos[pop1]
        color = self.cm_scalar_mappable.to_rgba(p)

        self.ax.annotate(
            "", xy=(to, t), xytext=(frm, t),
            arrowprops=dict(
                arrowstyle="-|>,head_width=.3,head_length=.6",
                fc=color, ec=color, ls=":", shrinkA=0,
                alpha=alpha,
                connectionstyle="arc3,rad={rad}".format(
                    rad=rad)))

        if pulse_label:
            if -rad*(frm - to) > 0:
                verticalalignment = "top"
            else:
                verticalalignment = "bottom"

            self.ax.annotate(
                "{}".format("{:.1f}%".format(100*p)),
                xy=(.5*(frm+to) + adj_label_x, t + adj_label_y),
                color="black", weight="bold",
                horizontalalignment="center",
                verticalalignment=verticalalignment)

    def get_N_legend_values(self):
        curr_N = 10**np.floor(
            np.log(min(self.all_N)) / np.log(10))
        N_legend_values = [curr_N]
        while curr_N < max(self.all_N):
            curr_N *= 4
            N_legend_values.append(curr_N)

        return N_legend_values

    def draw_N_legend(self, N_legend_values=None, title="N", **kwargs):
        """Draw legend of population sizes.

        ``**kwargs`` are passed onto :meth:`matplotlib.axes.Axes.legend`.

        :param N_legend_values: Values of ``N`` to plot.
        :param title: Title of legend
        :rtype: :class:`matplotlib.legend.Legend`
        """
        if N_legend_values is None:
            N_legend_values = self.get_N_legend_values()

        N_legend = co.OrderedDict(
            ("{:.1e}".format(N), mpl.lines.Line2D(
                [], [], linewidth=self.N_to_linewidth(N)))
            for N in sorted(N_legend_values))

        return self.ax.legend(
            N_legend.values(), N_legend.keys(),
            title=title, **kwargs)


class AdditionalDemographyPlot(DemographyPlot):
    def __init__(self, parent_plot, model):
        self.leafs = model.leafs
        self.x_pos = parent_plot.x_pos
        self.ax = parent_plot.ax
        self.base_N = parent_plot.base_N
        self.pmin = parent_plot.pmin
        self.pmax = parent_plot.pmax
        self.cm_scalar_mappable = parent_plot.cm_scalar_mappable
        self.additional_times = parent_plot.additional_times

        self._init_plot(model)

    def draw_frame(self, *a, **kw):
        pass

    def draw_xticks(self, *a, **kw):
        raise NotImplementedError

    def draw_pulse_colorbar(self, *a, **kw):
        raise NotImplementedError

    def draw_N_legend(self, *a, **kw):
        raise NotImplementedError


class _DemographyPlot(object):
    def __init__(self, x_pos_dict, default_N, additional_times):
        self.pop_lines = {
            p: PopulationLine(p, x, additional_times, default_N)
            for p, x in x_pos_dict.items()}
        self.pop_arrows = []

    def pop_to_t(self, pop, t, add_time=True):
        pop = self.pop_lines[pop]
        pop.goto_time(t, add_time=add_time)
        return pop

    def add_leaf(self, pop, t):
        self.pop_to_t(pop, t).active = True
        self.pop_to_t(pop, t).next_is_leaf = True

    def set_size(self, pop, t, N):
        pop = self.pop_to_t(pop, t, add_time=False)
        if pop.curr_N != N:
            pop.goto_time(t, add_time=True)
        pop.curr_N = N

    def set_growth(self, pop, t, g):
        pop = self.pop_to_t(pop, t, add_time=True)
        pop.curr_g = g

    def move_lineages(self, pop1, pop2, t, p, pulse_name=None):
        pop2 = self.pop_lines[pop2]
        #pop2.goto_time(t, add_time=not pop2.active)
        pop2.goto_time(t)
        pop2.active = True

        pop1 = self.pop_lines[pop1]
        #pop1.goto_time(t, add_time=(p == 1))
        pop1.goto_time(t)
        if p == 1:
            pop1.step_time(hq.heappop(pop1.time_stack))
            pop1.active = False
        self.pop_arrows.append(PopulationArrow(
            pop1, pop2, t, p, pop1.curr_N, pop2.curr_N, pulse_name))


PopulationPoint = co.namedtuple(
    "PopulationPoint", ["t", "N", "g", "is_leaf"])


PopulationArrow = co.namedtuple(
    "PopulationArrow", ["to_pop", "from_pop", "t", "p", "to_N", "from_N", "pulse_name"])


class PopulationLine(object):
    def __init__(self, name, x, times, N):
        self.name = name
        self.x = x
        if not times:
            self.time_stack = []
        else:
            self.time_stack = list(times)
        hq.heapify(self.time_stack)
        self.curr_N = N
        self.curr_g = 0.0
        self.curr_t = 0.0
        self.points = []
        self.active = False
        self.next_is_leaf = False

    def __repr__(self):
        return "{cls}(name={name}, x={x}) at {loc}".format(
            cls=type(self), name=self.name, x=self.x,
            loc=hex(id(self)))

    def step_time(self, nxt_t, add=True):
        assert self.curr_t <= nxt_t
        self.curr_N = self.curr_N * np.exp(
            -self.curr_g * (nxt_t - self.curr_t))
        self.curr_t = nxt_t

        if add and self.active:
            self.points.append(PopulationPoint(
                self.curr_t, self.curr_N, self.curr_g, self.next_is_leaf))
            self.next_is_leaf = False

    def _push_time(self, t):
        assert t >= self.curr_t
        if t not in self.time_stack:
            hq.heappush(self.time_stack, t)

    def goto_time(self, t, add_time=True):
        # if exponentially growing, add extra time points whenever
        # the population size doubles
        if self.curr_g != 0 and t < float('inf'):
            halflife = np.abs(np.log(.5) / self.curr_g)
            add_t = self.curr_t + halflife
            while add_t < t:
                self._push_time(add_t)
                add_t += halflife

        while self.time_stack and self.time_stack[0] < t:
            self.step_time(hq.heappop(self.time_stack))
        self.step_time(t, add=False)
        if add_time:
            # put t on queue to be added when processing next event
            # (allows further events to change population size before plotting)
            self._push_time(t)




import pytest
import momi
from momi import expected_sfs
import momi.likelihood
from demo_utils import simple_admixture_demo
import autograd.numpy as np
import itertools
import random
import sys
from collections import Counter
import logging

# test subsampling of SNPs


@pytest.mark.parametrize("fold,normalized",
                         ((random.choice((True, False)), random.choice((True, False))),))
def test_subconfigs(fold, normalized):
    demo = simple_admixture_demo()

    num_bases = 1000
    mu = 1.
    n_loci = 1000
    sampled_n_dict={"a":4,"b":5}
    sfs = demo.simulate_data(
        muts_per_gen=mu/num_bases,
        recoms_per_gen=0,
        length=num_bases,
        num_replicates=n_loci,
        sampled_n_dict=sampled_n_dict)._sfs

    if fold:
        configs = sfs.fold().configs
    else:
        configs = sfs.configs

    demo = demo._get_demo(sampled_n_dict)
    sub_idxs = np.array(random.sample(
        list(range(len(configs))), int(len(configs) / 2) + 1))
    assert len(sub_idxs) > 0 and len(sub_idxs) < len(configs)

    val1 = momi.expected_sfs(demo, configs,
                             normalized=normalized, folded=fold)[sub_idxs]

    sub_configs = momi.data.configurations._ConfigList_Subset(configs, sub_idxs)
    val2 = momi.expected_sfs(demo, sub_configs,
                             normalized=normalized, folded=fold)

    assert np.allclose(val1, val2)


#@pytest.mark.parametrize("fold,use_mut",
#                         ((random.choice((True, False)), random.choice((True, False))),))
#def test_subsfs(fold, use_mut):
#    demo = simple_admixture_demo()
#    demo.demo_hist = demo.demo_hist.rescaled()
#
#    n_loci = 10
#    #mut_rate = 100.
#    #sfs = momi.simulate_ms(scrm_path, demo.demo_hist,
#    #                       sampled_pops=demo.pops, sampled_n=demo.n,
#    #                       num_loci=n_loci, mut_rate=mut_rate).sfs
#    num_bases = 1000
#    mu = 100.
#    sfs = demo.demo_hist.simulate_data(
#        demo.pops, demo.n,
#        mutation_rate=mu/num_bases,
#        recombination_rate=0,
#        length=num_bases,
#        num_replicates=n_loci).sfs
#
#    if fold:
#        sfs = sfs.fold()
#
#    locus = random.choice(list(range(n_loci)))
#    #subsfs = momi.likelihood._SubSfs(sfs.configs, sfs._counts_ij[locus,:])
#    counts = np.zeros(sfs.n_nonzero_entries)
#    loc_idxs, loc_counts = sfs.loc_idxs[locus], sfs.loc_counts[locus]
#    counts[loc_idxs] = loc_counts
#    subsfs = momi.data.sfs._sub_sfs(sfs.configs, counts)
#
#    if not use_mut:
#        mut_rate = None
#
#    val1 = momi.likelihood._composite_log_likelihood(
#        sfs, demo.demo_hist, mut_rate=mut_rate, vector=True, folded=fold)[locus]
#    #val2 = momi.CompositeLogLikelihood(subsfs, mut_rate=mut_rate, folded=fold).evaluate(demo)
#    val2 = momi.likelihood._composite_log_likelihood(
#        subsfs, demo.demo_hist, mut_rate=mut_rate, folded=fold)
#
#    assert np.isclose(val1, val2)
#
#
#@pytest.mark.parametrize("fold",
#                         (random.choice((True, False)),))
#def test_subsfs2(fold):
#    demo = simple_admixture_demo()
#    demo.demo_hist = demo.demo_hist.rescaled()
#
#    n_loci = 10
#    #mut_rate = 100.
#    #sfs = momi.simulate_ms(scrm_path, demo.demo_hist,
#    #                       sampled_pops=demo.pops, sampled_n=demo.n,
#    #                       num_loci=n_loci, mut_rate=mut_rate).sfs
#    num_bases = 1000
#    mu = 100.
#    sfs = demo.demo_hist.simulate_data(
#        demo.pops, demo.n,
#        mutation_rate=mu/num_bases,
#        recombination_rate=0,
#        length=num_bases,
#        num_replicates=n_loci).sfs
#
#    if fold:
#        sfs = sfs.fold()
#
#    subsfs_list = momi.likelihood._subsfs_list(sfs, 10, np.random)
#    total = [([tuple(map(tuple, sfs.configs[i])) for i in subsfs.configs.sub_idxs],
#              subsfs._total_freqs)
#             for subsfs in subsfs_list]
#    total = [Counter(dict(list(zip(cnfs, cnts)))) for cnfs, cnts in total]
#    total = dict(sum(total, Counter()))
#
#    assert total == sfs.to_dict()
#
#
#@pytest.mark.parametrize("fold",
#                         (random.choice((True, False)),))
#def test_subliks(fold):
#    pre_demo_func = lambda *x: simple_admixture_demo(x)
#    demo_func = lambda *x: pre_demo_func(*x).demo_hist.rescaled()
#    rnd = np.random.RandomState()
#    x0 = rnd.normal(size=7)
#
#    demo = pre_demo_func(*x0)
#
#    n_loci = 100
#    #mut_rate = 10.
#    #sfs = momi.simulate_ms(scrm_path, demo.demo_hist.rescaled(),
#    #                       sampled_pops=demo.pops, sampled_n=demo.n,
#    #                       num_loci=n_loci, mut_rate=mut_rate).sfs
#    num_bases = 1000
#    mu = 10.
#    sfs = demo.demo_hist.rescaled().simulate_data(
#        demo.pops, demo.n,
#        mutation_rate=mu/num_bases,
#        recombination_rate=0,
#        length=num_bases,
#        num_replicates=n_loci).sfs
#
#    if fold:
#        sfs = sfs.fold()
#
#    n_chunks = 10
#
#    surface = momi.SfsLikelihoodSurface(
#        sfs, demo_func=demo_func, mut_rate=None, folded=fold)
#    subsurfaces = surface._get_stochastic_pieces(n_chunks, np.random)
#
#    val0 = [s.log_lik(x0) for s in subsurfaces]
#    val1 = surface.log_lik(x0)
#
#    assert np.isclose(np.sum(val0), val1)


#@pytest.mark.parametrize("folded,use_pairwise_diffs",
#                         #itertools.product((True, False), repeat=2))
#                         (tuple(random.choice((True, False)) for _ in range(2)),))
#def test_stochastic_inference(folded, use_pairwise_diffs):
#    num_runs = 1000
#    mu = 1.0
#    sampled_pops = (1, 2, 3)
#    sampled_n = (5, 5, 5)
#
#    def get_demo(t0, t1):
#        return momi.demographic_history([('-ej', t0, 1, 2), ('-ej', t0 + t1, 2, 3)])
#    true_x = np.array([.5, .2])
#    true_demo = get_demo(*true_x)
#
#    #sfs = momi.simulate_ms(ms_path, true_demo,
#    #                       sampled_pops=sampled_pops, sampled_n=sampled_n,
#    #                       num_loci=num_runs, mut_rate=mu).sfs
#    num_bases = 1000
#    sfs = true_demo.simulate_data(
#        sampled_pops, sampled_n,
#        mutation_rate=mu/num_bases,
#        recombination_rate=0,
#        length=num_bases,
#        num_replicates=num_runs).sfs
#
#    if folded:
#        sfs = sfs.fold()
#
#    log_prior = lambda x: np.sum(-x / true_x)
#
#    def callback(x):
#        if x.iteration % 10 == 0:
#            print(x.iteration, x.fun, x)
#    optimize_res = momi.SfsLikelihoodSurface(
#        sfs, demo_func=get_demo, mut_rate=mu, folded=folded,
#        log_prior=log_prior, use_pairwise_diffs=use_pairwise_diffs
#    )._stochastic_surfaces(n_minibatches=10).find_mle(
#        np.array([.1, .9]), bounds=[(1e-100, None), (1e-100, None)],
#        method="adam", svrg_epoch=10, num_iters=1000, callback=callback)
#    print(optimize_res)
#
#    inferred_x = optimize_res.x
#    error = (true_x - inferred_x) / true_x
#    print("# Truth:\n", true_x)
#    print("# Inferred:\n", inferred_x)
#    print("# Max Relative Error: %f" % max(abs(error)))
#    print("# Relative Error:", "\n", error)
#
#    assert max(abs(error)) < .1
#    #assert False

def test_stochastic_jointime_inference(
        sampled_n=(5, 5, 5), folded=False, add_n=0,
        use_theta=False, theta=.1, num_runs=10000):
    t0 = random.uniform(.25, 2.5)
    t1 = t0 + random.uniform(.5, 5.0)

    num_bases = 1e3
    theta = theta / num_bases

    model = momi.DemographicModel(1, muts_per_gen=theta)
    model.add_leaf(1)
    model.add_leaf(2)
    model.add_leaf(3)
    #model.add_parameter("join_time", t0, scaled_lower=0.0, scaled_upper=t1)
    #model.add_param("join_time", x0=t0, upper_x=t1)
    model.add_time_param("join_time", t0, upper=t1)
    model.move_lineages(1, 2, t="join_time")
    model.move_lineages(2, 3, t=t1)

    sampled_pops = (1, 2, 3)

    data = model.simulate_data(
        num_bases, 0, num_runs,
        sampled_n_dict=dict(zip(sampled_pops, sampled_n)))

    sfs = data.extract_sfs(1)
    assert sfs.n_snps() > 0
    sfs = sfs._copy(sampled_n=np.array(sampled_n) + add_n)
    if folded:
        sfs = sfs.fold()

    print((t0, t1))

    #prim_log_lik = momi.likelihood._raw_log_lik
    #prim_log_lik.reset_grad_count()
    #assert not prim_log_lik.num_grad_calls()

    if not use_theta:
        model.set_mut_rate(None)

    #model.set_x(random.uniform(0, t1), "join_time")
    model.set_params({"join_time": random.uniform(0,t1)}, scaled=True)

    model.set_data(sfs)
    res = model.stochastic_optimize(n_minibatches=10, num_iters=100, svrg_epoch=10)

    # make sure autograd is calling the rearranged gradient
    #assert bool(prim_log_lik.num_grad_calls())

    print(res.jac)
    #assert abs(res.x - t0) / t0 < .05
    #assert (model.get_params()["join_time"] - t0) / t0 < .05
    assert (res.parameters["join_time"] - t0) / t0 < .05


# def test_complex_stochastic_inference():
##     seed = np.random.randint(2**32-1)
# seed = 270543553
# seed = 2300087928
# seed = 3443948355
# seed = 2092909111
# seed = 2470901291
# seed = 2173658782
# print("SEED",seed)
# np.random.seed(seed)
##
##     true_N_chb_bottom, true_N_chb_top, true_pulse_t, true_pulse_p, true_ej_chb, true_ej_yri = 10.,.1,.25,.03,.5,1.5
# def demo_func(log_N_chb_bottom, log_N_chb_top, log_pulse_t, log_pulse_p, log_ej_chb, log_ej_yri):
##         pulse_t = 2**(log_pulse_t) * true_pulse_t
##         ej_chb = pulse_t + 2**(log_ej_chb) * (true_ej_chb - true_pulse_t)
##         ej_yri = ej_chb + 2**(log_ej_yri) * (true_ej_yri - true_ej_chb)
##
##         N_chb_top = 2**(log_N_chb_top) * true_N_chb_top
##         N_chb_bottom = 2**(log_N_chb_bottom) * true_N_chb_bottom
##
##         G_chb = -np.log(N_chb_top / N_chb_bottom) / ej_chb
##
# events = [('-en', 0., 'chb', N_chb_bottom),
##                   ('-eg', 0, 'chb' , G_chb),
##                   ('-ep', pulse_t, 'chb', 'nea', .25*(4.0*true_pulse_p)**((2**(log_pulse_p))**.5)),
##                   ('-ej', ej_chb, 'chb', 'yri'),
##                   ('-ej', ej_yri, 'yri', 'nea'),
# ]
# return momi.make_demography(events, ('yri','chb'), (14,10))
##
##     true_params = np.zeros(6)
##     true_demo = demo_func(*true_params)
##
# sfs = momi.simulate_ms(ms_path, true_demo,
# 1000, mut_rate=10.0,
# seeds=[np.random.randint(2**32-1) for _ in range(3)]).sfs
##
##
# define (lower,upper) bounds on the parameter space
##     lower_bounds = true_params + np.log(.01)/np.log(2)
##     upper_bounds = true_params + np.log(100.)/np.log(2)
##     bounds = list(zip(lower_bounds, upper_bounds))
##
# pick a random start value for the parameter search
##     start_params = np.array([np.random.uniform(l,h) for (l,h) in bounds])
##
# def callback(x):
# if x.iteration % 10 == 0:
##             print(x.iteration, x.fun, x)
# try:
# optimize_res = momi.SfsLikelihoodSurface(sfs, demo_func=demo_func, mut_rate=None).find_mle(start_params, bounds=bounds, method="tnc", maxiter=500, callback=callback)
# optimize_res = momi.SfsLikelihoodSurface(sfs, demo_func=demo_func, mut_rate=None).find_mle(start_params, bounds=bounds, method="L-BFGS-B", callback=callback)
# optimize_res = momi.SfsLikelihoodSurface(sfs, demo_func=demo_func, mut_rate=None)._stochastic_surfaces(n_minibatches=100, exact=50).find_mle(start_params, bounds=bounds, method="svrg", stepsize=.1, iter_per_epoch=10, max_epochs=100, callback=callback)
##         optimize_res = momi.SfsLikelihoodSurface(sfs, demo_func=demo_func, mut_rate=None)._stochastic_surfaces(n_minibatches=100, exact=50).find_mle(start_params, bounds=bounds, method="adam", svrg_epoch=10, num_iters=1000, callback=callback)
# except:
# print("SEED",seed)
# raise
##
##     true_x = true_params
##     inferred_x = optimize_res.x
##     error = inferred_x-true_x
# print("# Gradient:\n", optimize_res.jac)
# print("# Truth:\n", true_x)
# print("# Inferred:\n", inferred_x)
# print("# Ratio Inferred/Truth:","\n", 2.**error)
# print("# Max Log2 Ratio: %f" % max(abs(error)))
# try:
# print ("# Epochs: ", optimize_res.nepoch)
# except: pass
# print("SEED",seed)
# assert max(abs(np.log(error))) < .1



import collections as co
import autograd.numpy as np
import networkx as nx
import msprime
from .math_functions import hypergeom_quasi_inverse, binom_coeffs, _apply_error_matrices, convolve_trailing_axes, sum_trailing_antidiagonals
from .size_history import ConstantHistory, ExponentialHistory, PiecewiseHistory


# FIXME: we always assume default_N=1.0 for now
# NOTE sample_sizes should be an OrderedDict?
def _build_demo_graph(events, sample_sizes, params_dict, default_N):
    _G = nx.DiGraph()
    #_G.graph['event_cmds'] = tuple(events)
    _G.graph['default_N'] = default_N
    _G.graph['events_as_edges'] = []
    # the nodes currently at the root of the graph, as we build it up from the
    # leafs
    _G.graph['roots'] = {}

    for e in events:
        e.add_to_graph(_G, sample_sizes, params_dict)

    assert _G.nodes
    _G.graph['roots'] = [r for _, r in list(
        _G.graph['roots'].items()) if r is not None]

    if len(_G.graph['roots']) != 1:
        raise DemographyError("Must have a single root population")

    node, = _G.graph['roots']
    _set_sizes(_G.nodes[node], float('inf'))

    _G.graph['sampled_pops'] = tuple(sample_sizes.keys())
    _G.graph["events"] = tuple(events)
    _G.graph["params"] = co.OrderedDict(params_dict)
    return _G

class ParamsDict(co.OrderedDict):
    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __repr__(self):
        return self.__class__.__name__ + "({})".format(
            dict(self))

    def __dir__(self):
        return list(self.keys())


# TODO rename to ScaledParameter
class Parameter(object):
    def __init__(self, name, x0, transform_x,
                 inv_transform_x, x_bounds, rgen):
        self.name = name
        self.x = x0
        self.x_bounds = list(x_bounds)
        self.transform_x = transform_x
        self.inv_transform_x = inv_transform_x
        self.rgen = rgen

        # TODO some sort of test that inv_transform_x is actually the inverse of transform_x

    def copy(self):
        return Parameter(
            name=self.name, x0=self.x,
            transform_x=self.transform_x,
            inv_transform_x=self.inv_transform_x,
            x_bounds=self.x_bounds,
            rgen=self.rgen)

    def resample(self, params):
        self.x = self.inv_transform_x(self.rgen(params),
                                      params)

    def update_params_dict(self, params_dict, x=None):
        if x is None:
            x = self.x
        params_dict[self.name] = self.transform_x(
            x, params_dict)


def get_event_from_old(oldstyle_event):
    flag, t = oldstyle_event[:2]
    rest = oldstyle_event[2:]
    if flag=="-en":
        i,N = rest
        return SizeEvent(t,N,i,1.,.25)
    elif flag=="-eg":
        i,g = rest
        return GrowthEvent(t,g,i,1.,.25)
    elif flag=="-ej":
        i,j = rest
        return JoinEvent(t,i,j,1.,.25)
    elif flag=="-ep":
        i,j,p = rest
        return PulseEvent(t,p,i,j,1.,.25)
    elif flag=="-eSample":
        i,n = rest
        return LeafEvent(t,i,1.,.25)
    else:
        assert False

class LeafEvent(object):
    def __init__(self, t, pop, N_e, gen_time):
        self.t = TimeValue(t, N_e, gen_time)
        self.pop = pop

    def __repr__(self):
        return "LeafEvent(t={}, pop={})".format(
            self.t.x, self.pop)

    def add_to_plot(self, params_dict, demo_plot):
        demo_plot.add_leaf(self.pop, self.t(params_dict, scaled=False))

    def add_to_graph(self, G, sample_sizes, params_dict):
        t=self.t(params_dict)
        i=self.pop
        n=sample_sizes[self.pop]
        G.add_node((i, 0), lineages=n)

        if i in G.graph['roots']:
            if G.graph['roots'][i] is None:
                raise DemographyError(
                    "Invalid events: pop {0} removed by move_lineages before sampling time".format(i))

            #G.nodes[(i,0)]['model'] = _TrivialHistory()
            G.nodes[(i, 0)]['sizes'] = [
                {'t': t, 'N': G.graph['default_N'], 'growth_rate':None}]
            _set_sizes(G.nodes[(i, 0)], t)

            prev = G.graph['roots'][i]
            _set_sizes(G.nodes[prev], t)

            assert prev[0] == i and prev[1] != 0
            newpop = (i, prev[1] + 1)
            _ej_helper(G, t, (i, 0), prev, newpop)
        else:
            newpop = (i, 0)
            G.nodes[newpop]['sizes'] = [
                {'t': t, 'N': G.graph['default_N'], 'growth_rate':None}]
        G.graph['roots'][i] = newpop

    def get_msprime_event(self, params_dict, pop_ids_dict):
        return None

def _get_pop_id(pop, pop_ids_dict):
    if pop not in pop_ids_dict:
        pop_ids_dict[pop] = len(pop_ids_dict)
    return pop_ids_dict[pop]

class SizeEvent(object):
    def __init__(self, t, N, pop, N_e, gen_time):
        self.t = TimeValue(t, N_e, gen_time)
        self.N = SizeValue(N, N_e)
        self.pop = pop

    def oldstyle_event(self, prm_dict):
        return [("-en", self.t(prm_dict), self.pop, self.N(prm_dict))]

    def __repr__(self):
        return "SizeEvent(t={}, pop={}, N={})".format(
            self.t.x, self.pop, self.N.x)

    def add_to_plot(self, params_dict, demo_plot):
        demo_plot.set_size(self.pop, self.t(params_dict, scaled=False),
                           self.N(params_dict, scaled=False))

    def add_to_graph(self, G, sample_sizes, params_dict):
        t=self.t(params_dict)
        i=self.pop
        N=self.N(params_dict)
        _check_en_eg_pops(G, '-en', t, i, N)
        G.nodes[G.graph['roots'][i]]['sizes'].append(
            {'t': t, 'N': N, 'growth_rate': None})

    def get_msprime_event(self, params_dict, pop_ids_dict):
        t = self.t(params_dict)
        i = _get_pop_id(self.pop, pop_ids_dict)
        N = self.N(params_dict)
        return msprime.PopulationParametersChange(
            t, initial_size=N / 4, growth_rate=0, population_id=i)


class JoinEvent(object):
    def __init__(self, t, pop1, pop2, N_e, gen_time):
        self.t = TimeValue(t, N_e, gen_time)
        self.pop1 = pop1
        self.pop2 = pop2

    def __repr__(self):
        return "JoinEvent(t={}, pop1={}, pop2={})".format(
            self.t.x, self.pop1, self.pop2)

    def oldstyle_event(self, prm_dict):
        return [("-ej", self.t(prm_dict), self.pop1, self.pop2)]

    def add_to_plot(self, params_dict, demo_plot):
        demo_plot.move_lineages(self.pop1, self.pop2,
                                self.t(params_dict, scaled=False), 1)

    def add_to_graph(self, G, sample_sizes, params_dict):
        t=self.t(params_dict)
        i=self.pop1
        j=self.pop2
        if i not in G.graph['roots']:
            G.graph['roots'][i] = None
            # don't need to do anything else
            return
        _check_ej_ep_pops(G, '-ej', t, i, j)

        i0, j0 = (G.graph['roots'][k] for k in (i, j))
        j1 = (j, j0[1] + 1)
        assert j1 not in G.nodes()

        for k in i0, j0:
            # sets the TruncatedSizeHistory, and N_top and growth_rate for all
            # epochs
            _set_sizes(G.nodes[k], t)
        _ej_helper(G, t, i0, j0, j1)

        G.graph['roots'][j] = j1
        G.graph['roots'][i] = None

    def get_msprime_event(self, params_dict, pop_ids_dict):
        t = self.t(params_dict)
        i = _get_pop_id(self.pop1, pop_ids_dict)
        j = _get_pop_id(self.pop2, pop_ids_dict)
        return msprime.MassMigration(t, i, j)

class PulseEvent(object):
    def __init__(self, t, p, pop1, pop2, N_e, gen_time):
        self.t = TimeValue(t, N_e, gen_time)
        self.p = EventValue(p)
        self.pop1 = pop1
        self.pop2 = pop2

    def __repr__(self):
        s = "PulseEvent(t={}, pop1={}, pop2={}, p={})"
        return s.format(self.t.x, self.pop1, self.pop2,
                        self.p.x)

    def oldstyle_event(self, prm_dict):
        return [("-ep", self.t(prm_dict), self.pop1,
                 self.pop2, self.p(prm_dict))]

    def add_to_plot(self, params_dict, demo_plot):
        demo_plot.move_lineages(self.pop1, self.pop2,
                                self.t(params_dict, scaled=False),
                                self.p(params_dict, scaled=False),
                                pulse_name=self.p.x)

    def add_to_graph(self, G, sample_sizes, params_dict):
        t=self.t(params_dict)
        i=self.pop1
        j=self.pop2
        pij=self.p(params_dict)
        if pij < 0. or pij > 1.:
            raise DemographyError("Invalid pulse {0} from {1} to {2} at {3}: pulse probability must be between 0,1".format(pij, j, i, t))

        if i not in G.graph['roots']:
            # don't need to do anything
            return

        _check_ej_ep_pops(G, '-ep', t, i, j, pij)

        children = {k: G.graph['roots'][k] for k in (i, j)}
        for v in list(children.values()):
            _set_sizes(G.nodes[v], t)

        parents = {k: (v[0], v[1] + 1) for k, v in list(children.items())}
        assert all([par not in G.nodes for par in list(parents.values())])

        prev_sizes = {k: G.nodes[c]['sizes'][-1] for k, c in list(children.items())}
        for k, s in list(prev_sizes.items()):
            G.add_node(parents[k], sizes=[
                    {'t': t, 'N': s['N_top'], 'growth_rate':s['growth_rate']}])

        G.add_edge(parents[i], children[i], prob=1. - pij)
        G.add_edge(parents[j], children[i], prob=pij)
        G.add_edge(parents[j], children[j])

        new_event = tuple((parents[u], children[v])
                        for u, v in ((i, i), (j, i), (j, j))
                        )
        G.graph['events_as_edges'] += [new_event]

        for k, v in list(parents.items()):
            G.graph['roots'][k] = v

    def get_msprime_event(self, params_dict, pop_ids_dict):
        t = self.t(params_dict)
        i = _get_pop_id(self.pop1, pop_ids_dict)
        j = _get_pop_id(self.pop2, pop_ids_dict)
        pij=self.p(params_dict)

        return msprime.MassMigration(
            t, i, j, proportion=pij)

class GrowthEvent(object):
    def __init__(self, t, g, pop, N_e, gen_time):
        self.t = TimeValue(t, N_e, gen_time)
        self.pop = pop
        self.g = RateValue(g, N_e, gen_time)

    def __repr__(self):
        return "GrowthEvent(t={}, pop={}, g={})".format(
            self.t.x, self.pop, self.g.x)

    def oldstyle_event(self, prm_dict):
        return [("-eg", self.t(prm_dict), self.pop,
                 self.g(prm_dict))]

    def add_to_plot(self, params_dict, demo_plot):
        demo_plot.set_growth(self.pop, self.t(params_dict, scaled=False),
                             self.g(params_dict, scaled=False))

    def add_to_graph(self, G, sample_sizes, params_dict):
        t=self.t(params_dict)
        i=self.pop
        growth_rate=self.g(params_dict)
        _check_en_eg_pops(G, '-eg', t, i, growth_rate)
        G.nodes[G.graph['roots'][i]]['sizes'].append(
            {'t': t, 'growth_rate': growth_rate})

    def get_msprime_event(self, params_dict, pop_ids_dict):
        t = self.t(params_dict)
        i = _get_pop_id(self.pop, pop_ids_dict)
        g = self.g(params_dict)
        return msprime.PopulationParametersChange(
            t, growth_rate=g, population_id=i)

## helper functions for building demo.
## TODO remove/rename these!!

def _ej_helper(G, t, i0, j0, j1):
    prev = G.nodes[j0]['sizes'][-1]
    G.add_node(j1, sizes=[{'t': t, 'N': prev['N_top'],
                           'growth_rate':prev['growth_rate']}])

    new_edges = ((j1, i0), (j1, j0))
    G.graph['events_as_edges'].append(new_edges)
    G.add_edges_from(new_edges)


def _check_en_eg_pops(G, *event):
    flag, t, i = event[:3]
    if i in G.graph['roots'] and G.graph['roots'][i] is None:
        raise DemographyError(
            "Invalid set_size event at time {0}: pop {1} was already removed by previous move_lineages".format(t, i))

    if i not in G.graph['roots']:
        G.graph['roots'][i] = (i, 1)
        G.add_node(G.graph['roots'][i],
                   sizes=[{'t': t, 'N': G.graph['default_N'], 'growth_rate':None}],
                   )


def _check_ej_ep_pops(G, *event):
    flag, t, i, j = event[:4]
    for k in (i, j):
        if k in G.graph['roots'] and G.graph['roots'][k] is None:
            raise DemographyError(
                "Invalid move_lineages event at time {0}: pop {1} was already removed by previous move_lineages".format(t, k))

    if j not in G.graph['roots']:
        G.graph['roots'][j] = (j, 1)
        G.add_node(G.graph['roots'][j],
                   sizes=[{'t': t, 'N': G.graph['default_N'], 'growth_rate':None}],
                   )


def _set_sizes(node_data, end_time):
    assert 'model' not in node_data

    # add 'model_func' to node_data, add information to node_data['sizes']
    sizes = node_data['sizes']
    # add a dummy epoch with the end time
    sizes.append({'t': end_time})

    # do some processing
    N, growth_rate = sizes[0]['N'], sizes[0]['growth_rate']
    pieces = []
    for i in range(len(sizes) - 1):
        sizes[i]['tau'] = tau = (sizes[i + 1]['t'] - sizes[i]['t'])

        if 'N' not in sizes[i]:
            sizes[i]['N'] = N
        if 'growth_rate' not in sizes[i]:
            sizes[i]['growth_rate'] = growth_rate
        growth_rate = sizes[i]['growth_rate']
        N = sizes[i]['N']

        if growth_rate is not None and tau != float('inf'):
            pieces.append(ExponentialHistory(
                tau=tau, growth_rate=growth_rate, N_bottom=N))
            N = pieces[-1].N_top
        else:
            if growth_rate != 0. and growth_rate is not None and tau == float('inf'):
                raise DemographyError("Final epoch must have 0 growth rate")
            pieces.append(ConstantHistory(tau=tau, N=N))

        sizes[i]['N_top'] = N

        if not all([sizes[i][x] >= 0.0 for x in ('tau', 'N', 'N_top')]):
            raise DemographyError("Negative time or population size in {sizes}".format(
                sizes=[{k: str(v) for k, v in s.items()} for s in sizes]))
    sizes.pop()  # remove the final dummy epoch

    assert len(pieces) > 0
    if len(pieces) == 0:
        node_data['model'] = pieces[0]
    else:
        node_data['model'] = PiecewiseHistory(pieces)


class DemographyError(Exception):
    pass


## represent values as float, str, or lambda

class EventValue(object):
    def __init__(self, x):
        self.x = x
        self.scale = 1.0

    def __call__(self, params_dict, scaled=True):
        if isinstance(self.x, str):
            x = params_dict[self.x]
        elif callable(self.x):
            x = self.x(params_dict)
        else:
            x = self.x
        if scaled:
            return x / self.scale
        else:
            return x


class SizeValue(EventValue):
    def __init__(self, N, N_e):
        self.x = N
        self.scale = N_e


class TimeValue(EventValue):
    def __init__(self, t, N_e, gen_time):
        self.x = t
        self.scale = 4.0 * N_e * gen_time


class RateValue(EventValue):
    def __init__(self, r, N_e, gen_time):
        self.x = r
        self.scale = .25 / N_e / gen_time





import pytest

import numpy as np

from momi import expected_sfs_tensor_prod, expected_tmrca
from demo_utils import simple_admixture_demo, simple_five_pop_demo


def check_tmrca(demo):
    tmrca = expected_tmrca(demo)
    print(tmrca)
    for i in range(len(demo.sampled_pops)):
        vecs = [np.ones(n + 1) for n in demo.sampled_n]
        vecs[0] = np.arange(len(vecs[0])) / (len(vecs[0]) - 1.0)
        tmrca2 = expected_sfs_tensor_prod(
            vecs, demo, sampled_pops=demo.sampled_pops)
        print(tmrca2)
        assert np.isclose(tmrca, tmrca2)


def test_five_pop_tmrca():
    check_tmrca(simple_five_pop_demo()._get_demo(dict(zip(range(1,6), [10]*5))))


def test_admixture_tmrca():
    check_tmrca(simple_admixture_demo()._get_demo({"a":3,"b":6}))



import itertools as it
import collections as co
import json
import autograd.numpy as np
import numpy as raw_np
from cached_property import cached_property
import scipy
import scipy.sparse
import gzip
import os
import logging
from .compressed_counts import _hashed2config, _config2hashable
from .compressed_counts import CompressedAlleleCounts
from .configurations import ConfigList
from .configurations import _ConfigList_Subset
from ..util import memoize_instance


def site_freq_spectrum(sampled_pops, freqs_by_locus, length=None):
    """Create :class:`Sfs` from a list of dicts of counts.

    ``freqs_by_locus[l][config]`` gives the frequency of ``config`` \
    on locus ``l``.

    A ``config`` is a ``(n_pops,2)``-array, where ``config[p][a]`` \
    is the count of allele ``a`` in population ``p``. ``a=0`` represents \
    the ancestral allele while ``a=1`` represents the derived allele.

    :param list sampled_pops: list of population names
    :param list freqs_by_locus: list of dict
    :param float,None length: Number of bases in the dataset. Set to ``None`` if unknown.
    :rtype: :class:`Sfs`
    """
    index2loc = []
    index2count = []
    loci_counters = []

    def chained_sequences():
        for loc, locus_iter in enumerate(freqs_by_locus):
            loci_counters.append(co.Counter())
            try:
                locus_iter = locus_iter.items()
            except AttributeError:
                locus_iter = zip(locus_iter, iter(lambda: 1, 0))
            for config, count in locus_iter:
                index2loc.append(loc)
                index2count.append(count)
                yield config

    compressed_counts = CompressedAlleleCounts.from_iter(
        chained_sequences(), len(sampled_pops))
    config_array = compressed_counts.config_array
    index2uniq = compressed_counts.index2uniq

    assert len(index2loc) == len(index2count) and len(
        index2count) == len(index2uniq)
    for loc, count, uniq in zip(index2loc, index2count, index2uniq):
        loci_counters[loc][uniq] += count

    configs = ConfigList(sampled_pops, config_array, None, None)
    return Sfs(loci_counters, configs, folded=False, length=length)



class Sfs(object):
    """
    Class for representing observed site frequnecy spectrum data.

    To create a :class:`Sfs` object, use \
    :meth:`SnpAlleleCounts.extract_sfs`, \
    :meth:`Sfs.load`, or :func:`site_freq_spectrum`. \
    Do NOT call the class constructor directly, it is for internal use only.
    """
    @classmethod
    def from_matrix(cls, mat, configs, *args, **kwargs):
        # convert to csc
        mat = scipy.sparse.csc_matrix(mat)
        indptr = mat.indptr
        loci = []
        for i in range(mat.shape[1]):
            loci.append(np.array([
                mat.indices[indptr[i]:indptr[i+1]],
                mat.data[indptr[i]:indptr[i+1]]]))

        return cls(loci, configs, *args, **kwargs)

    @classmethod
    def load(cls, f):
        """Load :class:`Sfs` from file created by :meth:`Sfs.dump` or ``python -m momi.extract_sfs``

        :param str,file f: file object or file name to read in
        :rtype: :class:`Sfs`
        """
        if isinstance(f, str):
            fname = os.path.expanduser(f)
            if fname.endswith(".gz"):
                with gzip.open(fname, "rt") as f:
                    return cls.load(f)
            else:
                with open(fname) as f:
                    return cls.load(f)

        logging.getLogger(__name__).info("Reading json...")
        info = json.load(f)
        logging.getLogger(__name__).info("Finished reading json...")

        logging.getLogger(__name__).info("Constructing configs...")
        configs = info.pop("configs")
        logging.getLogger(__name__).info("{} unique configs detected".format(len(configs)))
        # don't use autograd here for performance
        configs = raw_np.array(configs, dtype=int)
        logging.getLogger(__name__).info("Converted to numpy array...")
        configs = ConfigList(info.pop("sampled_pops"), configs)
        logging.getLogger(__name__).info("Finished constructing configs")

        logging.getLogger(__name__).info("Constructing SFS...")
        loci = []
        for locus, locus_rows in it.groupby(
                info.pop("(locus,config_id,count)"), lambda x: x[0]):
            loci.append({config_id: count
                        for _, config_id, count in locus_rows})

        ret = cls(loci, configs, **info)
        logging.getLogger(__name__).info("Finished constructing SFS")

        return ret

    def __init__(self, loci, configs, folded, length):
        self.folded = folded
        self._length = length

        self.configs = configs

        self.loc_idxs, self.loc_counts = [], []
        for loc in loci:
            if len(loc) == 0:
                self.loc_idxs.append(np.array([], dtype=int))
                self.loc_counts.append(np.array([], dtype=float))
            else:
                try:
                    loc.items()
                except:
                    loc = np.array(loc)
                    if len(loc.shape) == 2:
                        assert loc.shape[0] == 2
                        idxs, cnts = loc[0, :], loc[1, :]
                    else:
                        idxs, cnts = np.unique(loc, return_counts=True)
                else:
                    idxs, cnts = zip(*loc.items())
                self.loc_idxs.append(np.array(idxs, dtype=int))
                self.loc_counts.append(np.array(cnts, dtype=float))

        if len(self.loc_idxs) > 1:
            self._total_freqs = self.freqs_matrix.dot(np.ones(self.n_loci))
            assert self._total_freqs.shape == (self.freqs_matrix.shape[0],)
        else:
            # avoid costly building of frequency matrix, when there are many
            # Sfs's of a single locus (e.g. in many stochastic minibatches)
            idxs, = self.loc_idxs
            cnts, = self.loc_counts
            self._total_freqs = np.zeros(len(self.configs))
            self._total_freqs[idxs] = cnts

        assert not np.any(self._total_freqs == 0)

    def dump(self, f):
        """Write Sfs to file

        :param str,file f: Filename or object. If name ends with ".gz" gzip it.
        """
        if isinstance(f, str):
            fname = os.path.expanduser(f)
            if fname.endswith(".gz"):
                with gzip.open(fname, "wt") as f:
                    self.dump(f)
            else:
                with open(fname, 'w') as f:
                    self.dump(f)
            return

        print("{", file=f)
        print('\t"sampled_pops": {},'.format(
            json.dumps(list(self.sampled_pops))), file=f)
        print('\t"folded": {},'.format(
            json.dumps(self.folded)), file=f)
        print('\t"length": {},'.format(
            json.dumps(self._length)), file=f)
        print('\t"configs": [', file=f)
        n_configs = len(self.configs.value)
        for i, c in enumerate(self.configs.value.tolist()):
            if i != n_configs - 1:
                print("\t\t{},".format(c), file=f)
            else:
                # no trailing comma
                print("\t\t{}".format(c), file=f)
        print("\t],", file=f)
        print('\t"(locus,config_id,count)": [', file=f)
        for loc in range(self.n_loci):
            n_loc_rows = len(self.loc_idxs[loc])
            assert n_loc_rows == len(self.loc_counts[loc])
            for i in range(n_loc_rows):
                config_id = int(self.loc_idxs[loc][i])
                count = float(self.loc_counts[loc][i])
                if loc == self.n_loci - 1 and i == n_loc_rows - 1:
                    # no trailing comma
                    print("\t\t{}".format(json.dumps(
                        (loc, config_id, count))), file=f)
                else:
                    print("\t\t{},".format(json.dumps(
                        (loc, config_id, count))), file=f)
        print("\t]", file=f)
        print("}", file=f)

    @property
    def populations(self):
        return self.sampled_pops

    @memoize_instance
    def combine_loci(self):
        # return copy with all loci combined
        return self.from_matrix(self.freqs_matrix.sum(axis=1),
                                self.configs, self.folded,
                                self._length)

    @property
    def freqs_matrix(self):
        """Sparse matrix representing the frequencies at each locus.

        The ``(i,j)``-th entry gives the frequency of the ``i``-th config in :attr:`Sfs.config_array` at locus ``j``

        :rtype: :class:`scipy.sparse.csr_matrix`
        """
        return self.csr_freqs_matrix

    @cached_property
    def csr_freqs_matrix(self):
        return _csr_freq_matrix_from_counters(
            self.loc_idxs, self.loc_counts, len(self.configs))

    @cached_property
    def avg_pairwise_hets(self):
        # avg number of hets per ind per pop (assuming Hardy-Weinberg)
        n_nonmissing = np.sum(self.configs.value, axis=2)
        # for denominator, assume 1 allele is drawn from whole sample, and 1
        # allele is drawn only from nomissing alleles
        denoms = np.maximum(n_nonmissing * (self.sampled_n - 1), 1.0)
        p_het = 2 * self.configs.value[:, :, 0] * \
            self.configs.value[:, :, 1] / denoms

        return self.freqs_matrix.T.dot(p_het)

    def resample(self):
        """Create a new SFS by resampling blocks with replacement.

        Note the resampled SFS is assumed to have the same length in base pairs \
        as the original SFS, which may be a poor assumption if the blocks are not of equal length.

        :returns: Resampled SFS
        :rtype: :class:`Sfs`
        """
        loci = np.random.choice(
            np.arange(self.n_loci), size=self.n_loci, replace=True)
        mat = self.freqs_matrix[:, loci]
        to_keep = np.asarray(mat.sum(axis=1) > 0).squeeze()
        to_keep = np.arange(len(self.configs))[to_keep]
        mat = mat[to_keep, :]
        configs = _ConfigList_Subset(self.configs, to_keep)

        return self.from_matrix(mat, configs, self.folded, self.length)

    @property
    def sampled_n(self):
        """Number of samples per population

        ``sampled_n[i]`` is the number of haploids \
        in the ``i``-th population of :attr:`Sfs.sampled_pops`

        :returns: 1-d array of integers
        :rtype: :class:`numpy.ndarray`
        """
        return self.configs.sampled_n

    @property
    def ascertainment_pop(self):
        return self.configs.ascertainment_pop

    @property
    def sampled_pops(self):
        """Names of sampled populations

        :returns: 1-d array of strings
        :rtype: :class:`numpy.ndarray`
        """
        return self.configs.sampled_pops

    @property
    def n_loci(self):
        """The number of loci in the dataset

        :rtype: int
        """
        return len(self.loc_idxs)

    @property
    def n_nonzero_entries(self):
        return len(self.configs)

    @property
    def config_array(self):
        """Array of unique configs.

        This is a 3-d array with shape ``(n_configs, n_pops, 2)``. \
        ``config_array[i, p, a]`` gives the count of allele ``a`` \
        in population ``p`` in configuration ``i``. \
        Allele ``a=0`` is ancestral, ``a=1`` is derived. \
        :attr:`Sfs.sampled_pops` gives the ordering of the populations.

        :returns: array of configs
        :rtype: :class:`numpy.ndarray`
        """
        return self.configs.config_array

    @property
    def length(self):
        """Length of data in bases. ``None`` if unknown.

        :rtype: float
        """
        return self._length

    @memoize_instance
    def n_snps(self, vector=False):
        """Number of SNPs, either per-locus or overall.

        :param bool vector: Whether to return a single total count, or an array of counts per-locus
        """
        if vector:
            return raw_np.array(self.freqs_matrix.sum(axis=0)).reshape((-1,))
        else:
            return np.sum(self._total_freqs)

    def __eq__(self, other):
        loci_dicts = self._get_dict(vector=True)
        try:
            return loci_dicts == other._get_dict(vector=True)
        except AttributeError:
            return False

    def _get_dict(self, vector=False, locus=None):
        if vector:
            assert locus is None
            return [self._get_dict(locus=loc) for loc in range(self.n_loci)]
        elif locus is None:
            return dict(zip(map(_config2hashable, self.configs),
                            self._total_freqs))
        return dict(zip(
            (_config2hashable(self.configs[i]) for i in self.loc_idxs[locus]),
            self.loc_counts[locus]))

    def to_dict(self, vector=False):
        if not vector:
            return {_hashed2config(k): v for k, v in self._get_dict().items()}
        else:
            return [{_hashed2config(k): v for k, v in d.items()}
                    for d in self._get_dict(vector=True)]

    # FIXME: This method returns the _negative_ entropy, and should be renamed.
    @cached_property
    def _entropy(self):
        counts = self._total_freqs
        n_snps = float(self.n_snps())
        p = counts / n_snps
        # return np.sum(p * np.log(p))
        ret = np.sum(p * np.log(p))

        # correct for missing data
        sampled_n = np.sum(self.configs.value, axis=2)
        sampled_n_counts = co.Counter()
        assert len(counts) == len(sampled_n)
        for c, n in zip(counts, sampled_n):
            n = tuple(n)
            sampled_n_counts[n] += c
        sampled_n_counts = np.array(
            list(sampled_n_counts.values()), dtype=float)

        ret = ret + np.sum(sampled_n_counts / n_snps *
                           np.log(n_snps / sampled_n_counts))
        assert not np.isnan(ret)
        return ret

    def _get_muts_poisson_entropy(self, use_pairwise_diffs):
        if use_pairwise_diffs:
            return self._pairwise_muts_poisson_entropy
        else:
            return self._total_muts_poisson_entropy

    @cached_property
    def _total_muts_poisson_entropy(self):
        lambd = self.n_snps(vector=True)
        lambd = lambd[lambd > 0]
        return np.sum(-lambd + lambd * np.log(lambd)
                      - scipy.special.gammaln(lambd + 1))

    @cached_property
    def _pairwise_muts_poisson_entropy(self):
        lambd = self.avg_pairwise_hets
        ret = -lambd + lambd * np.log(lambd) - scipy.special.gammaln(lambd + 1)
        ret[lambd <= 1e-16] = 0
        ret = ret * self.sampled_n / float(np.sum(
            self.sampled_n[self.ascertainment_pop]))
        return np.sum(ret[:, self.ascertainment_pop])

    def fold(self):
        """
        :returns: A copy of the SFS, but with folded entries.
        :rtype: :class:`Sfs`
        """
        def get_folded(config):
            if tuple(config[:, 0]) < tuple(config[:, 1]):
                return config[:, ::-1]
            else:
                return config

        compressed_folded = CompressedAlleleCounts.from_iter(
            map(get_folded, self.configs),
            len(self.sampled_pops), sort=False)

        return self.from_matrix(
            compressed_folded.index2uniq_mat @ self.freqs_matrix,
            ConfigList(self.sampled_pops, compressed_folded.config_array,
                        sampled_n=self.sampled_n,
                        ascertainment_pop=self.ascertainment_pop),
            folded=True, length=self._length)

    def _copy(self, sampled_n=None):
        """
        See also: ConfigList._copy()
        """
        if sampled_n is None:
            sampled_n = self.sampled_n
        return self.from_matrix(
            self.csr_freqs_matrix,
            ConfigList(self.sampled_pops, self.configs.value,
                        sampled_n=sampled_n,
                        ascertainment_pop=self.ascertainment_pop),
            self.folded, self._length)

    def _integrate_sfs(self, weights, vector=False, locus=None):
        if vector:
            assert locus is None
            return np.array([self._integrate_sfs(weights, locus=loc)
                             for loc in range(self.n_loci)])
        if locus is None:
            idxs, counts = slice(None), self._total_freqs
        else:
            idxs, counts = self.loc_idxs[locus], self.loc_counts[locus]
        return np.sum(weights[idxs] * counts)

    def subset_populations(self, populations, non_ascertained_pops=None):
        """Returns lower-dimensional SFS restricted to a subset of populations.

        :param list populations: list of populations to subset to
        :param list non_ascertained_pops: list of populations to treat as non-ascertained (SNPs must be polymorphic with respect to the ascertained populations)
        :returns: Lower-dimensional SFS restricted to a subset of populations.
        :rtype: :class:`Sfs`
        """
        if non_ascertained_pops is None:
            non_ascertained_pops = []
            for pop, asc in zip(self.sampled_pops, self.ascertainment_pop):
                if not asc:
                    non_ascertained_pops.append(pop)
        return self._subset_populations(tuple(populations),
                                        tuple(non_ascertained_pops))

    @memoize_instance
    def _subset_populations(self, populations, non_ascertained_pops):
        # old indexes of the new population ordering
        old_sampled_pops = list(self.sampled_pops)
        old_pop_idx = np.array([
            old_sampled_pops.index(p) for p in populations], dtype=int)

        # old ascertainment
        ascertained = dict(zip(self.sampled_pops,
                               self.ascertainment_pop))
        # restrict to new populations
        ascertained = {p: ascertained[p] for p in populations}
        # previously non-ascertained pops should remain so
        for pop, is_asc in ascertained.items():
            if not is_asc:
                assert pop in non_ascertained_pops
        # update ascertainment
        for pop in non_ascertained_pops:
            ascertained[pop] = False
        # convert from dict to numpy array
        ascertained = np.array([ascertained[p] for p in populations])

        # keep only polymorphic configs
        asc_only = self.configs[:, old_pop_idx[ascertained], :]
        asc_is_poly = (asc_only.sum(axis=1) != 0).all(axis=1)
        asc_is_poly = np.arange(len(asc_is_poly))[asc_is_poly]
        sub_sfs = self._subset_configs(asc_is_poly)

        # get the new configs
        new_configs = CompressedAlleleCounts.from_iter(
            sub_sfs.configs[:, old_pop_idx, :],
            len(populations), sort=False)

        return self.from_matrix(
            new_configs.index2uniq_mat @ sub_sfs.freqs_matrix,
            ConfigList(populations, new_configs.config_array,
                        ascertainment_pop=ascertained),
            self.folded, self._length)

    def _subset_configs(self, idxs):
        return self.from_matrix(
            self.csr_freqs_matrix[idxs, :],
            _ConfigList_Subset(self.configs, idxs),
            self.folded, self._length)

    @property
    def sfs(self):
        return self

    @cached_property
    def p_missing(self):
        """Estimate of probability that a random allele from each population is missing.

        Missingness is estimated as follows: \
        from each SNP remove a random allele; \
        if the resulting config is monomorphic, then ignore. \
        If polymorphic, then count whether the removed allele \
        is missing or not.

        This avoids bias from fact that we don't observe \
        some polymorphic configs that appear monomorphic \
        after removing the missing alleles.

        :returns: 1-d array of missingness per population
        :rtype: :class:`numpy.ndarray`
        """
        counts = self._total_freqs
        sampled_n = self.sampled_n
        n_pops = len(self.sampled_pops)

        config_arr = self.configs.value
        # augment config_arr to contain the missing counts
        n_miss = sampled_n - np.sum(config_arr, axis=2)
        config_arr = np.concatenate((config_arr, np.reshape(
            n_miss, list(n_miss.shape)+[1])), axis=2)

        ret = []
        for i in range(n_pops):
            n_valid = []
            for allele in (0, 1, -1):
                # configs with removed allele
                removed = np.array(config_arr)
                removed[:, i, allele] -= 1
                # is the resulting config polymorphic?
                valid_removed = (removed[:, i, allele] >= 0) & np.all(
                    np.sum(removed[:, :, :2], axis=1) > 0, axis=1)

                # sum up the valid configs
                n_valid.append(np.sum(
                    (counts * config_arr[:, i, allele])[valid_removed]))
            # fraction of valid configs with missing additional allele
            ret.append(n_valid[-1] / float(sum(n_valid)))
        return np.array(ret)


def _csr_freq_matrix_from_counters(idxs_by_loc, cnts_by_loc,
                                   n_configs):
    data = []
    indices = []
    indptr = []
    n_loc = 0
    for idxs, cnts in zip(idxs_by_loc, cnts_by_loc):
        indptr.append(len(data))
        data.extend(cnts)
        indices.extend(idxs)
        n_loc += 1
    indptr.append(len(data))
    return scipy.sparse.csc_matrix((data, indices, indptr),
                                   shape=(n_configs, n_loc)).tocsr()


def _get_subsample_counts(configs, n):
    subconfigs, weights = [], []
    for pop_comb in it.combinations_with_replacement(configs.sampled_pops, n):
        subsample_n = co.Counter(pop_comb)
        subsample_n = np.array([subsample_n[pop]
                                for pop in configs.sampled_pops], dtype=int)
        if np.any(subsample_n > configs.sampled_n):
            continue

        for sfs_entry in it.product(*(range(sub_n + 1)
                                      for sub_n in subsample_n)):
            sfs_entry = np.array(sfs_entry, dtype=int)
            if np.all(sfs_entry == 0) or np.all(sfs_entry == subsample_n):
                # monomorphic
                continue

            sfs_entry = np.transpose([subsample_n - sfs_entry, sfs_entry])
            cnt_vec = configs.subsample_probs(sfs_entry)
            if not np.all(cnt_vec == 0):
                subconfigs.append(sfs_entry)
                weights.append(cnt_vec)

    return np.array(subconfigs), np.array(weights)


class SimpleNamespace(object):
    pass



import pytest
import random
import autograd.numpy as np
from momi import SfsLikelihoodSurface
import momi

def test_archaic_and_pairwisediffs():
    #logging.basicConfig(level=logging.DEBUG)
    theta = 1
    N_e = 1.0
    join_time = 1.0
    num_runs = 1000

    def logit(p):
        return np.log(p / (1. - p))

    def expit(x):
        return 1. / (1. + np.exp(-x))

    n_bases = 1000
    model = momi.DemographicModel(
        N_e, muts_per_gen=theta/4./N_e/n_bases)

    model.add_time_param(
        "sample_t", random.uniform(0.001, join_time-.001) / join_time,
        upper=join_time)
    model.add_size_param("N", 1.0)
    model.add_leaf("a", N="N")
    model.add_leaf("b", t="sample_t", N="N")
    model.move_lineages("a", "b", join_time)

    data = model.simulate_data(length=n_bases,
                               recoms_per_gen=0,
                               num_replicates=num_runs,
                               sampled_n_dict={"a": 2, "b": 2})

    model.set_data(data.extract_sfs(1),
                   use_pairwise_diffs=False,
                   mem_chunk_size=-1)

    true_params = np.array(list(model.get_params().values()))
    #model.set_x([logit(random.uniform(.001, join_time-.001) / join_time),
    model.set_params([
        logit(random.uniform(.001, join_time-.001) / join_time),
        random.uniform(-1, 1)],
                     scaled=True)
    res = model.optimize(method="trust-ncg", hessp=True)
    inferred_params = np.array(list(model.get_params().values()))

    assert np.max(np.abs(np.log(true_params / inferred_params))) < .2


@pytest.mark.parametrize("folded,add_n",
                         ((f, n) for f in (True, False) for n in (0, 3)))
def test_jointime_inference(folded, add_n):
    check_jointime_inference(folded=folded, add_n=add_n)

def test_ascertainment_inference():
    check_jointime_inference(non_ascertained_pops=[3], use_theta=True,
                             sampled_n=(5,5,1))

def check_jointime_inference(
        sampled_n=(5, 5, 5), folded=False, add_n=0,
        use_theta=False, theta=.1, num_runs=10000,
        non_ascertained_pops=None):
    t0 = random.uniform(.25, 2.5)
    t1 = t0 + random.uniform(.5, 5.0)

    num_bases = 1e3
    theta = theta / num_bases

    model = momi.DemographicModel(1, muts_per_gen=theta)
    model.add_leaf(1)
    model.add_leaf(2)
    model.add_leaf(3)
    #model.add_parameter("join_time", t0, scaled_lower=0.0, scaled_upper=t1)
    #model.add_param("join_time", x0=t0, upper_x=t1)
    model.add_time_param("join_time", t0, upper=t1)
    model.move_lineages(1, 2, t="join_time")
    model.move_lineages(2, 3, t=t1)

    sampled_pops = (1, 2, 3)

    data = model.simulate_data(
        num_bases, 0, num_runs,
        sampled_n_dict=dict(zip(sampled_pops, sampled_n)))

    sfs = data.extract_sfs(1)
    assert sfs.n_snps() > 0
    sfs = sfs._copy(sampled_n=np.array(sampled_n) + add_n)
    if folded:
        sfs = sfs.fold()

    print((t0, t1))

    #prim_log_lik = momi.likelihood._raw_log_lik
    #prim_log_lik.reset_grad_count()
    #assert not prim_log_lik.num_grad_calls()

    if not use_theta:
        model.set_mut_rate(None)

    #model.set_x(random.uniform(0, t1), "join_time")
    model.set_params({"join_time": random.uniform(0,t1)}, scaled=True)

    model.set_data(sfs, non_ascertained_pops=non_ascertained_pops)
    res = model.optimize()

    # make sure autograd is calling the rearranged gradient
    #assert bool(prim_log_lik.num_grad_calls())

    print(res.jac)
    #assert abs(res.x - t0) / t0 < .05
    #assert (model.get_params()["join_time"] - t0) / t0 < .05
    assert (res.parameters["join_time"] - t0) / t0 < .05


@pytest.mark.parametrize("folded", (True, False))
def test_underflow_robustness(folded):
    num_runs = 1000
    sampled_pops = (1, 2, 3)
    sampled_n = (5, 5, 5)

    n_bases = int(1e3)
    demo = momi.DemographicModel(1.0, .25, muts_per_gen=2.5 / n_bases)
    for p in sampled_pops:
        demo.add_leaf(p)
    demo.add_time_param("t0")
    demo.add_time_param("t1", lower_constraints=["t0"])
    demo.move_lineages(1, 2, "t0")
    demo.move_lineages(2, 3, "t1")

    true_params = np.array([0.5, 0.7])
    demo.set_params(true_params)

    data = demo.simulate_data(
        length=n_bases,
        recoms_per_gen=0.0,
        num_replicates=num_runs,
        sampled_n_dict=dict(zip(sampled_pops, sampled_n)))

    sfs = data.extract_sfs(1)
    if folded:
        sfs = sfs.fold()

    demo.set_data(sfs)
    demo.set_params({"t0": 0.1, "t1": 100.0})
    optimize_res = demo.optimize()

    print(optimize_res)
    inferred_params = np.array(list(demo.get_params().values()))

    error = (true_params - inferred_params) / true_params
    print("# Truth:\n", true_params)
    print("# Inferred:\n", inferred_params)
    print("# Max Relative Error: %f" % max(abs(error)))
    print("# Relative Error:", "\n", error)

    assert max(abs(error)) < .1




import pytest
import momi
from momi import expected_sfs
import momi.likelihood
from demo_utils import simple_admixture_demo, simple_five_pop_demo
import autograd.numpy as np
import itertools
import random
from collections import Counter
import scipy

#def test_subsample_inds():
#    demo = simple_admixture_demo()
#    #demo.demo_hist = demo.demo_hist.rescaled()
#    #data = momi.simulate_ms(ms_path, demo.demo_hist,
#    #                        sampled_pops=demo.pops,
#    #                        sampled_n=demo.n,
#    #                        num_loci=1000, mut_rate=1.0)
#    num_bases = 1000
#    mu = 1.0
#    num_replicates = 1000
#    data = demo.simulate_data(
#        length=num_bases, recoms_per_gen=0,
#        muts_per_gen=1.0 / num_bases,
#        num_replicates=num_replicates,
#        sampled_n_dict={"b":2,"a":3})
#    #data = demo.demo_hist.simulate_data(
#    #    demo.pops, demo.n,
#    #    mutation_rate=mu/num_bases,
#    #    recombination_rate=0,
#    #    length=num_bases,
#    #    num_replicates=num_replicates)
#    assert data.sfs.n_snps() > 0
#    assert data.subsample_inds(4).sfs == data.sfs.subsample_inds(4)


def test_count_subsets():
    demo = simple_admixture_demo()
    #data = momi.simulate_ms(ms_path, demo.demo_hist,
    #                        sampled_pops=demo.pops,
    #                        sampled_n=demo.n,
    #                        num_loci=1000, mut_rate=1.0)
    num_bases = 1000
    mu = 1.0
    num_replicates = 100
    data = demo.simulate_data(
        muts_per_gen=mu/num_bases,
        recoms_per_gen=0,
        length=num_bases,
        num_replicates=num_replicates,
        sampled_n_dict={"b":2,"a":3}).extract_sfs(None)

    subconfig = []
    for n in data.sampled_n:
        sub_n = random.randrange(n+1)
        d = random.randrange(sub_n+1)
        subconfig.append([sub_n-d, d])

    subconfig_probs = np.ones(len(data.configs))
    for i, (a, d) in enumerate(subconfig):
        #subconfig_probs *= scipy.special.comb(
        #    data.configs.value[:, i, :], [a, d]).prod(axis=1)
        #subconfig_probs /= scipy.special.comb(
        #    data.configs.value[:, i, :].sum(axis=1), a+d)
        subconfig_probs *= scipy.stats.hypergeom.pmf(
            d, data.configs.value[:, i, :].sum(axis=1),
            data.configs.value[:, i, 1], a+d)

    assert np.allclose(subconfig_probs,
                       data.configs.subsample_probs(subconfig))


@pytest.mark.parametrize("folded,n_lins",
                         ((f, n) for f in (True, False) for n in ((2, 3), (0, 3))))
def test_simple_admixture_subsampling(folded, n_lins):
    check_subsampling(simple_admixture_demo(), {"b":2,"a":3},
                      3, folded=folded)


def check_subsampling(demo, sampled_n_dict, add_n, folded=False):
    config_list = momi.data.configurations.build_full_config_list(*zip(
        *sampled_n_dict.items()))
    if folded:
        config_list = momi.site_freq_spectrum(
            config_list.sampled_pops, [config_list]).fold().configs

    sfs1 = demo.expected_sfs(config_list, normalized=True, folded=folded)
    #sfs1 = expected_sfs(demo._get_demo(sampled_n_dict), config_list,
    #                    normalized=True, folded=folded, **kwargs)

    configs2 = config_list._copy(sampled_n=config_list.sampled_n + add_n)
    sfs2 = demo.expected_sfs(configs2, normalized=True, folded=folded)
    #sfs2 = expected_sfs(demo._get_demo(dict(zip(configs2.sampled_pops, configs2.sampled_n))),
    #                    configs2,
    #                    normalized=True, folded=folded, **kwargs)

    sfs1 = np.array(list(sfs1.values()))
    sfs2 = np.array(list(sfs2.values()))
    assert np.allclose(sfs1, sfs2)

    ## check sums to 1 even with error matrix
    #error_matrices = [np.exp(np.random.randn(n + 1, n + 1))
    #                  for n in configs2.sampled_n]
    #error_matrices = [
    #    np.einsum('ij,j->ij', x, 1. / np.sum(x, axis=0)) for x in error_matrices]

    #sfs3 = demo.expected_sfs(configs2, normalized=True, folded=folded,
    #                         error_matrices=error_matrices)
    #assert np.isclose(sum(sfs3.values()), 1.0)


def test_subsample_pops():
    demo = simple_five_pop_demo()
    num_bases = 1000
    mu = 1.0
    num_replicates = 100
    data = demo.simulate_data(
        muts_per_gen=mu/num_bases,
        recoms_per_gen=0,
        length=num_bases,
        num_replicates=num_replicates,
        sampled_n_dict={i: 5 for i in range(1,6)})

    sfs1 = data.subset_populations([1,2,3], [3]).extract_sfs(None)
    sfs2 = data.extract_sfs(None).subset_populations([1,2,3], [3])
    assert sfs1 == sfs2



"""Runnable module to convert VCF file to :class:`SnpAlleleCounts`.

Run this from the command line like ``python -m momi.read_vcf ...``. \
See the ``--help`` flag for command line options. \
Use :meth:`SnpAlleleCounts.read_vcf` to read the VCF from within Python.
"""

import argparse
import sys
import logging
from .data.snps import SnpAlleleCounts


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("vcf_file", help="VCF file to read. The VCF file should be indexed, e.g. with tabix or bcftools. Alternatively, use \"-\" to read the VCF from stdin.")
    parser.add_argument("ind2pop",
                        help="File whose first column is individual ID and second column is population ID")
    parser.add_argument("out_file", help="Output file to store counts. If ends with .gz, gzip it.")
    parser.add_argument("--bed", help="Mask file specifying which regions to read. Also used to determine the size of the data in bases. If not provided then user will need to manually specify the length when required. Do NOT use the same BED file across multiple VCFs or the length of those regions will be double-counted!")
    parser.add_argument("--no_aa", action='store_true',
                        help="Ignore AA information entirely; use folded SFS downstream.")
    parser.add_argument("--outgroup", default=None,
                        help="Set this population as outgroup to determine ancestral allele, instead of using the AA info field. Note the outgroup will not appear in the created data (as it always has allele 0).")
    parser.add_argument("--verbose", action="store_true")
    parser.add_argument("--info_aa_field", default="AA", help="INFO field to read ancestral allele from. Default is AA. Has no effect if --outgroup or --no_aa are set.")
    args = parser.parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.INFO, format="%(asctime)s %(message)s")


    with open(args.ind2pop) as f:
        ind2pop = dict([l.split() for l in f])

    if args.no_aa:
        ancestral_alleles = False
    elif args.outgroup:
        ancestral_alleles = args.outgroup
    else:
        ancestral_alleles = True

    SnpAlleleCounts.read_vcf(
        args.vcf_file, ind2pop, bed_file=args.bed,
        ancestral_alleles=ancestral_alleles,
        info_aa_field=args.info_aa_field).dump(args.out_file)




from .util import memoize, check_probs_matrix
from .math_functions import par_einsum
import scipy.sparse
import autograd.numpy as np
from autograd.numpy import dot, diag, exp


def moran_transition(t, n):
    assert t >= 0.0
    P, d, Pinv = moran_eigensystem(n)
    D = diag(exp(t * d))
    return check_probs_matrix(dot(P, dot(D, Pinv)))

def moran_action(t, v, axis=0):
    if v.shape[axis] == 1:
        return v

    n = v.shape[axis] - 1
    PDPinv = moran_transition(t, n)

    vlen, output_dim = len(v.shape), list(range(len(v.shape)))
    output_dim[axis] = vlen
    ret = par_einsum(v, list(range(vlen)), PDPinv, [vlen, axis], output_dim)
    assert ret.shape == v.shape
    return ret


@memoize
def rate_matrix(n, sparse_format="csr"):
    i = np.arange(n + 1)
    diag = i * (n - i) / 2.
    diags = [diag[:-1], -2 * diag, diag[1:]]
    M = scipy.sparse.diags(
        diags, [1, 0, -1], (n + 1, n + 1), format=sparse_format)
    return M


@memoize
def moran_eigensystem(n):
    M = rate_matrix(n).toarray()
    d, P = np.linalg.eig(M)
    return P, d, np.linalg.inv(P)




import pytest

import numpy as np

import momi
from momi import expected_sfs_tensor_prod, expected_total_branch_len
from demo_utils import simple_admixture_demo
from momi.math_functions import hypergeom_quasi_inverse

import autograd


def test_constructor():
    pre_demo = simple_admixture_demo()
    demo = pre_demo._get_demo({"b":2,"a":3})
    demo2 = momi.demography.Demography(
        demo._get_graph_structure(), demo._get_differentiable_part())

    assert np.allclose(expected_total_branch_len(demo),
                       expected_total_branch_len(demo2))

    # make sure it fails if we don't pass in the array values
    demo3 = momi.demography.Demography(demo._get_graph_structure())
    try:
        expected_total_branch_len(demo3)
    except:
        return
    assert False


class NoLookdownDemography(momi.demography.Demography):
    def __init__(self, demo):
        super(NoLookdownDemography, self).__init__(demo._G)

    def _n_at_node(self, node):
        if node[0] in self.sampled_pops and node[1] == 0:
            return self._G.nodes[node]['lineages']
        return np.sum(self._n_at_node(l) for l in self._G[node])


def test_pseudoinverse():
    demo0 = simple_admixture_demo()._get_demo({"b":2,"a":3})
    demo1 = NoLookdownDemography(demo0)

    p = 20
    vecs = [np.random.normal(size=(p, n + 1)) for n in demo0.sampled_n]

    vals0, vals1 = [expected_sfs_tensor_prod(vecs, d)
                    for d in (demo0, demo1)]

    assert np.allclose(vals0, vals1)

    n_lins_diff = np.array(
        [demo0._n_at_node(v) - demo1._n_at_node(v) for v in demo0._G])
    assert all(n_lins_diff <= 0)
    assert any(n_lins_diff < 0)


def test_hypergeom_pinv_eye():
    i = np.random.randint(2, 50)
    assert np.allclose(hypergeom_quasi_inverse(i, i),
                       np.eye(i + 1, i + 1))


def test_P():
    t1 = np.random.exponential(.25)
    t2 = np.random.exponential(.25) + t1
    t3 = np.random.exponential(.5) + t2
    p1 = np.random.uniform(0, 1)
    p2 = np.random.uniform(0, 1)

    i = np.random.choice([0, 1])
    j = 1 - i

    demo0 = momi.DemographicModel(1.0, .25)
    demo1 = momi.DemographicModel(1.0, .25)

    for d in (demo0, demo1):
        d.add_leaf(0)
        d.add_leaf(1)
        d.move_lineages(0, 1, t3)
    demo0.move_lineages(0, 1, t=t1, p=p1)
    demo0.move_lineages(i, j, t=t2, p=p2)

    demo1.move_lineages(0, 'x', t=t1, p=p1)
    demo1.move_lineages('x', 1, t=t1)
    demo1.move_lineages(i, 'y', t=t2, p=p2)
    demo1.move_lineages('y', j, t=t2)

    demo0 = demo0._get_demo({0:5,1:6})
    demo1 = demo1._get_demo({0:5,1:6})

    #root_event = ('-ej', t3, 0, 1)
    #pulse_events0 = [('-ep', t1, 0, 1, p1),
    #                 ('-ep', t2, i, j, p2)]
    #pulse_events1 = [('-ep', t1, 0, 'x', p1), ('-ej', t1, 'x', 1),
    #                 ('-ep', t2, i, 'y', p2), ('-ej', t2, 'y', j)]

    #demo0 = make_demography(pulse_events0 + [root_event],
    #                        (0, 1), (5, 6))
    #demo1 = make_demography(pulse_events1 + [root_event],
    #                        (0, 1), (5, 6))

    p = 20
    vecs = [np.random.normal(size=(p, n + 1)) for n in demo0.sampled_n]

    vals0, vals1 = [expected_sfs_tensor_prod(vecs, d)
                    for d in (demo0, demo1)]

    assert np.allclose(vals0, vals1)


def test_events_before_sample():
    n_events = 4
    t = [0.0]
    for i in range(n_events):
        t += [np.random.exponential(1. / float(n_events)) + t[-1]]
    t = t[1:]


    demo0 = momi.DemographicModel(1.0, .25)
    demo1 = momi.DemographicModel(1.0, .25)
    p = np.random.uniform(0, 1)
    for d in (demo0, demo1):
        d.add_leaf("a")
        d.add_leaf("b", t=t[3])
        d.move_lineages("a", "b", t=t[0], p=p)

    demo0.set_size("c", t=0, N=10, g=1)
    demo0.move_lineages("a", "c", t=t[1])
    demo0.move_lineages("c", "b", t=t[2])

    demo1.set_size("a", t=t[1], N=10*np.exp(-t[1]), g=1.0)
    demo1.move_lineages("a", "b", t=t[2])

    demo0, demo1 = [d._get_demo({"a":7,"b":5}) for d in (demo0, demo1)]

    #events = [('-ep', t[0], 'a', 'b', np.random.uniform(0, 1))]
    #demo0 = make_demography(events + [('-en', 0.0, 'c', 10.0), ('-eg', 0.0, 'c', 1.0),
    #                                  ('-ej', t[1], 'a', 'c'),
    #                                  ('-ej', t[2], 'c', 'b')],
    #                        sampled_pops=('a', 'b'), sampled_n=(7, 5),
    #                        sampled_t=(0., t[3]))

    #demo1 = make_demography(events + [('-en', t[1], 'a', 10.0 * np.exp(-t[1])), ('-eg', t[1], 'a', 1.0),
    #                                  ('-ej', t[2], 'a', 'b')],
    #                        sampled_pops=('a', 'b'), sampled_n=(7, 5),
    #                        sampled_t=(0., t[3]))

    vecs = [np.random.normal(size=(10, n + 1)) for n in demo0.sampled_n]
    val0, val1 = [expected_sfs_tensor_prod(vecs, d) for d in (demo0, demo1)]

    assert np.allclose(val0, val1)



from cached_property import cached_property
# autograd.numpy.array() inefficient, so use vanilla numpy here
import numpy as np
import scipy.sparse


# to hash configs, represent it as a str
# (this seems to be more memory efficient than representing it as a tuple)
def _config2hashable(config):
    return "\t".join("%d,%d" % (a, d) for a, d in config)


# the inverse function of _config2hashable
def _hashed2config(config_str):
    return tuple((int(a), int(d))
                 for a, d in (x.split(",")
                              for x in config_str.strip().split()))


class _CompressedList(object):
    def __init__(self, items=None):
        self.uniq_values = []
        self.value2uniq = {}
        self.index2uniq = []
        self._array = None

        if items:
            for i in items:
                self.append(i)

    def __eq__(self, other):
        return list(self) == list(other)

    def __len__(self):
        return len(self.index2uniq)

    def __getitem__(self, index):
        if isinstance(index, slice) or isinstance(index, np.ndarray):
            ret = _CompressedList()
            ret.value2uniq = dict(self.value2uniq)
            ret.uniq_values = list(self.uniq_values)
            ret.index2uniq = self._get_array()[index]
            return ret
        else:
            return self.uniq_values[self.index2uniq[index]]

    def _get_array(self):
        if self._array is None:
            self._array = np.array(self.index2uniq, dtype=int)
        return self._array

    def append(self, value):
        self._array = None
        try:
            uniq_idx = self.value2uniq[value]
        except KeyError:
            uniq_idx = len(self.uniq_values)
            self.value2uniq[value] = uniq_idx
            self.uniq_values.append(value)
        self.index2uniq.append(uniq_idx)

    def extend(self, iterator):
        for i in iterator:
            self.append(i)


class _CompressedHashedCounts(object):
    def __init__(self, npops):
        self.compressed_list = _CompressedList()
        self.npops = npops

    def __len__(self):
        return len(self.compressed_list)

    def append(self, config):
        self.compressed_list.append(_config2hashable(config))

    def index2uniq(self, i=None):
        if i is None:
            return self.compressed_list.index2uniq
        else:
            return self.compressed_list.index2uniq[i]

    def config_array(self):
        ret = np.zeros((len(self.compressed_list.uniq_values), self.npops, 2),
                       dtype=int)
        for i, config_str in enumerate(self.compressed_list.uniq_values):
            ret[i, :, :] = _hashed2config(config_str)
        return ret

    def compressed_allele_counts(self):
        return CompressedAlleleCounts(self.config_array(),
                                      self.index2uniq())


class CompressedAlleleCounts(object):
    @classmethod
    def from_iter(cls, config_iter, npops, sort=True):
        compressed_hashes = _CompressedHashedCounts(npops)
        for config in config_iter:
            compressed_hashes.append(config)
        return cls(compressed_hashes.config_array(),
                   compressed_hashes.index2uniq(),
                   sort=sort)

    def __init__(self, config_array, index2uniq,
                 sort=True):
        self.config_array = config_array
        self.index2uniq = np.array(index2uniq, dtype=int)
        if sort:
            self.sort_configs()

    def __getitem__(self, i):
        return self.config_array[self.index2uniq[i], :, :]

    def __eq__(self, other):
        try:
            return (np.all(self.config_array == other.config_array)
                    and np.all(self.index2uniq == other.index2uniq))
        except AttributeError:
            return False

    def __len__(self):
        return len(self.index2uniq)

    def filter(self, idxs):
        to_keep = self.index2uniq[idxs]
        uniq_to_keep, uniq_to_keep_inverse = np.unique(
            to_keep, return_inverse=True)
        return CompressedAlleleCounts(self.config_array[uniq_to_keep, :, :],
                                      uniq_to_keep_inverse)

    def sort_configs(self):
        # sort configs so that "(very) similar" configs are next to each other
        # and will end up in the same batch,
        # thus avoiding redundant computation
        # "similar" == configs have same num missing alleles
        # "very similar" == configs are folded copies of each other
        a = self.config_array[:, :, 0]  # ancestral counts
        d = self.config_array[:, :, 1]  # derived counts
        n = a + d  # totals

        n = list(map(tuple, n))
        a = list(map(tuple, a))
        d = list(map(tuple, d))

        folded = list(map(min, list(zip(a, d))))

        keys = list(zip(n, folded))
        sorted_idxs = sorted(range(len(n)), key=lambda i: keys[i])
        sorted_idxs = np.array(sorted_idxs, dtype=int)

        unsorted_idxs = [None] * len(sorted_idxs)
        for i, j in enumerate(sorted_idxs):
            unsorted_idxs[j] = i
        unsorted_idxs = np.array(unsorted_idxs, dtype=int)

        self.config_array = self.config_array[sorted_idxs, :, :]
        self.index2uniq = unsorted_idxs[self.index2uniq]

    def count_configs(self):
        return np.bincount(self.index2uniq, minlength=len(self.config_array))

    @cached_property
    def n_samples(self):
        return np.max(np.sum(self.config_array, axis=2), axis=0)

    @property
    def index2uniq_mat(self):
        return scipy.sparse.coo_matrix(
            (np.ones(len(self)), (self.index2uniq, np.arange(len(self)))),
            shape=(len(self.config_array), len(self)))



import pytest
import os
import random
import pickle
import functools
import autograd.numpy as np

import momi

from demo_utils import simple_three_pop_demo, simple_nea_admixture_demo

import scipy

# TODO: get these tests working if we reactivate dead Godambe code

#def check_cov(method, params, demo_func, num_runs, theta, bounds=None, subsample_inds=False, p_missing=None, **kwargs):
#    true_demo = demo_func(*params)
#    #seg_sites = simulate_ms(scrm_path, true_demo.demo_hist,
#    #                        sampled_pops=true_demo.pops, sampled_n=true_demo.n,
#    #                        num_loci=num_runs, mut_rate=theta,
#    #                        additional_ms_params="-r %f 1000" % theta)
#
#    # TODO fix this test after removing SegSites object
#
#    if p_missing:
#        seg_sites = momi.data.seg_sites._randomly_drop_alleles(
#            seg_sites, p_missing)
#
#    if subsample_inds:
#        seg_sites = seg_sites.subsample_inds(subsample_inds)
#        old_demo_func = demo_func
#        demo_func = functools.partial(
#            old_demo_func, sampled_n=seg_sites.sampled_n)
#
#    cmle_search_res = momi.SfsLikelihoodSurface(seg_sites, lambda *x: demo_func(
#        *x).demo_hist).find_mle(params, options={'maxiter': 1000}, bounds=bounds, **kwargs)
#    est_params = cmle_search_res.x
#
#    cr = momi.ConfidenceRegion(
#        est_params, lambda *x: demo_func(*x).demo_hist, seg_sites, regime=method, **kwargs)
#    cov = cr.godambe(inverse=True)
#
#    #print( cr.test(params,sims=100) )
#    #print( cr.test([params,est_params],sims=100) )
#    print(cr.test(params))
#    print(cr.test([params, est_params]))
#    return cr
#
#
#def check_jointime_cov(method, num_runs, theta):
#    t0 = random.uniform(.25, 2.5)
#    t1 = t0 + random.uniform(.5, 5.0)
#
#    def demo_func(t):
#        return simple_three_pop_demo(t, t1)
#    return check_cov(method, [t0], demo_func, num_runs, theta)
#
#
#def test_jointime_cov_many():
#    check_jointime_cov("many", 1000, 1.)
#
#
#def test_jointime_cov_long():
#    check_jointime_cov("long", 10, 100.)
#
#
#def check_admixture_cov(method, num_runs, theta, **kwargs):
#    return check_cov(method, simple_nea_admixture_demo.true_params, simple_nea_admixture_demo, num_runs, theta, bounds=simple_nea_admixture_demo.bounds, **kwargs)
#
#
#def test_admixture_cov_many():
#    check_admixture_cov("many", 1000, 1.)
#
#
#def test_admixture_cov_many_subsample():
#    check_admixture_cov("many", 1000, 1., subsample_inds=4)
#
#
#def test_admixture_cov_long():
#    check_admixture_cov("long", 5, 200.)
#
#
#def test_admixture_cov_long_subsample():
#    cr = check_admixture_cov("long", 2, 1000., subsample_inds=6, p_missing=.1)
#
#    # commented because pickling confidence region is too large in filesize
#    # test that pickling the confidence region works
#    #fname = "cr.tmp.pickle"
#    # try:
#    #    with open(fname,"wb") as f:
#    #        pickle.dump(cr, f)
#
#    #    with open(fname, "rb") as f:
#    #        cr2 = pickle.load(f)
#
#    #    assert cr.data == cr2.data
#    #    assert np.all(cr.score == cr2.score)
#    #    assert np.all(cr.score_cov == cr2.score_cov)
#    #    assert np.all(cr.fisher == cr2.fisher)
#    #    assert np.allclose(cr.godambe(), cr2.godambe())
#    # except:
#    #    if os.path.exists(fname):
#    #        os.remove(fname)
#    #    raise
#    # else:
#    #    os.remove(fname)
